# AUTO-GENERATED FILE. Do not edit. Run chalkpy stubgen to generate.
# fmt: off
# isort: skip_file
from __future__ import annotations

from chalk.features import DataFrame as DataFrame
from chalk.features import Features as Features
from chalk.features import Tags as Tags
from chalk.features.feature_set import FeaturesMeta as FeaturesMeta
from chalk.streams._windows import Windowed as Windowed
from chalk.utils.duration import Duration as Duration
from datetime import date as __stubgen_datetime_date
from datetime import datetime as __stubgen_datetime_datetime
from tests.client.test_client_serialization import Color as __stubgen_tests_client_test__client__serialization_Color
from tests.client.test_expand_features import Ankle as __stubgen_tests_client_test__expand__features_Ankle
from tests.client.test_expand_features import Foot as __stubgen_tests_client_test__expand__features_Foot
from tests.features.test_chained_feature_time import Homeowner as __stubgen_tests_features_test__chained__feature__time_Homeowner
from tests.features.test_chained_has_one import ExampleFraudOrg as __stubgen_tests_features_test__chained__has__one_ExampleFraudOrg
from tests.features.test_chained_has_one import ExampleFraudUser as __stubgen_tests_features_test__chained__has__one_ExampleFraudUser
from tests.features.test_df import Foo as __stubgen_tests_features_test__df_Foo
from tests.features.test_df import PydanticFoo as __stubgen_tests_features_test__df_PydanticFoo
from tests.features.test_df import Topping as __stubgen_tests_features_test__df_Topping
from tests.features.test_df import ToppingPrice as __stubgen_tests_features_test__df_ToppingPrice
from tests.features.test_features import SingleChildFS as __stubgen_tests_features_test__features_SingleChildFS
from tests.features.test_features import SingleParentFS as __stubgen_tests_features_test__features_SingleParentFS
from tests.features.test_iter import NoFunFeatures as __stubgen_tests_features_test__iter_NoFunFeatures
from tests.serialization.test_codec import Color as __stubgen_tests_serialization_test__codec_Color
from tests.serialization.test_codec import CustomClass as __stubgen_tests_serialization_test__codec_CustomClass
from tests.serialization.test_codec import MyDataclass as __stubgen_tests_serialization_test__codec_MyDataclass
from tests.sql.test_execute import EnumFeature as __stubgen_tests_sql_test__execute_EnumFeature
from tests.sql.test_execute import SQLExecuteNestedFeatures as __stubgen_tests_sql_test__execute_SQLExecuteNestedFeatures
from tests.streams.test_window import ChildWindow as __stubgen_tests_streams_test__window_ChildWindow
from tests.streams.test_window import GrandchildWindow as __stubgen_tests_streams_test__window_GrandchildWindow
from typing import Optional as Optional
from typing import Protocol as Protocol
from typing import Type as Type
from typing import Union as Union
from typing import overload as overload

class TacoMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def user_id(self) -> Type[str]: ...

    @property
    def price(self) -> Type[int]: ...

    @property
    def maybe_price(self) -> Type[Union[int, None]]: ...

    @property
    def maybe_str(self) -> Type[Union[str, None]]: ...

    @property
    def hat(self) -> Type[str]: ...

    @property
    def topping_id(self) -> Type[str]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def unzoned_ts(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def nicknames(self) -> Type[DataFrame]: ...

    @property
    def topping(self) -> Type[__stubgen_tests_features_test__df_Topping]: ...

    @property
    def foo(self) -> Type[__stubgen_tests_features_test__df_Foo]: ...

    @property
    def pydantic_foo(self) -> Type[__stubgen_tests_features_test__df_PydanticFoo]: ...

    @property
    def foos(self) -> Type[list[__stubgen_tests_features_test__df_Foo]]: ...

    @property
    def float_list(self) -> Type[list[float]]: ...

class Taco(Features, metaclass=TacoMetaclass):
    def __init__(
        self,
        id: str = ...,
        user_id: str = ...,
        price: int = ...,
        maybe_price: Union[int, None] = ...,
        maybe_str: Union[str, None] = ...,
        hat: str = ...,
        topping_id: str = ...,
        ts: __stubgen_datetime_datetime = ...,
        unzoned_ts: __stubgen_datetime_datetime = ...,
        nicknames: DataFrame = ...,
        topping: __stubgen_tests_features_test__df_Topping = ...,
        foo: __stubgen_tests_features_test__df_Foo = ...,
        pydantic_foo: __stubgen_tests_features_test__df_PydanticFoo = ...,
        foos: list[__stubgen_tests_features_test__df_Foo] = ...,
        float_list: list[float] = ...,
    ):
        self.id: str
        self.user_id: str
        self.price: int
        self.maybe_price: Union[int, None]
        self.maybe_str: Union[str, None]
        self.hat: str
        self.topping_id: str
        self.ts: __stubgen_datetime_datetime
        self.unzoned_ts: __stubgen_datetime_datetime
        self.nicknames: DataFrame
        self.topping: __stubgen_tests_features_test__df_Topping
        self.foo: __stubgen_tests_features_test__df_Foo
        self.pydantic_foo: __stubgen_tests_features_test__df_PydanticFoo
        self.foos: list[__stubgen_tests_features_test__df_Foo]
        self.float_list: list[float]

class TacoProtocol(Protocol):
    id: str
    user_id: str
    price: int
    maybe_price: Union[int, None]
    maybe_str: Union[str, None]
    hat: str
    topping_id: str
    ts: __stubgen_datetime_datetime
    unzoned_ts: __stubgen_datetime_datetime
    nicknames: DataFrame
    topping: __stubgen_tests_features_test__df_Topping
    foo: __stubgen_tests_features_test__df_Foo
    pydantic_foo: __stubgen_tests_features_test__df_PydanticFoo
    foos: list[__stubgen_tests_features_test__df_Foo]
    float_list: list[float]

class WowFSMetaclass(FeaturesMeta):
    @property
    def something(self) -> Type[str]: ...

    @property
    def something_else(self) -> Type[str]: ...

    @property
    def nocomment(self) -> Type[str]: ...

    @property
    def nope(self) -> Type[str]: ...

    @property
    def assigned(self) -> Type[str]: ...

    @property
    def bizarre(self) -> Type[str]: ...

    @property
    def goofy(self) -> Type[str]: ...

    @property
    def assigned_comment(self) -> Type[str]: ...

    @property
    def explicit(self) -> Type[str]: ...

    @property
    def assigned_comment_multiline(self) -> Type[str]: ...

    @property
    def time(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def nope_nope(self) -> Type[__stubgen_datetime_datetime]: ...

class WowFS(Features, metaclass=WowFSMetaclass):
    def __init__(
        self,
        something: str = ...,
        something_else: str = ...,
        nocomment: str = ...,
        nope: str = ...,
        assigned: str = ...,
        bizarre: str = ...,
        goofy: str = ...,
        assigned_comment: str = ...,
        explicit: str = ...,
        assigned_comment_multiline: str = ...,
        time: __stubgen_datetime_datetime = ...,
        nope_nope: __stubgen_datetime_datetime = ...,
    ):
        self.something: str
        self.something_else: str
        self.nocomment: str
        self.nope: str
        self.assigned: str
        self.bizarre: str
        self.goofy: str
        self.assigned_comment: str
        self.explicit: str
        self.assigned_comment_multiline: str
        self.time: __stubgen_datetime_datetime
        self.nope_nope: __stubgen_datetime_datetime

class WowFSProtocol(Protocol):
    something: str
    something_else: str
    nocomment: str
    nope: str
    assigned: str
    bizarre: str
    goofy: str
    assigned_comment: str
    explicit: str
    assigned_comment_multiline: str
    time: __stubgen_datetime_datetime
    nope_nope: __stubgen_datetime_datetime

class StreamFeaturesWindowMetaclass(FeaturesMeta):
    @property
    def uid(self) -> Type[str]: ...

    @property
    def scalar_feature__600__(self) -> Type[str]: ...

    @property
    def scalar_feature__1200__(self) -> Type[str]: ...

    @property
    def scalar_feature(self) -> Type[Windowed[str]]: ...

    @property
    def scalar_feature_2__600__(self) -> Type[str]: ...

    @property
    def scalar_feature_2__1200__(self) -> Type[str]: ...

    @property
    def scalar_feature_2(self) -> Type[Windowed[str]]: ...

    @property
    def basic(self) -> Type[str]: ...

    @property
    def child_id(self) -> Type[str]: ...

    @property
    def child(self) -> Type[__stubgen_tests_streams_test__window_ChildWindow]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class StreamFeaturesWindow(Features, metaclass=StreamFeaturesWindowMetaclass):
    def __init__(
        self,
        uid: str = ...,
        scalar_feature__600__: str = ...,
        scalar_feature__1200__: str = ...,
        scalar_feature: Windowed[str] = ...,
        scalar_feature_2__600__: str = ...,
        scalar_feature_2__1200__: str = ...,
        scalar_feature_2: Windowed[str] = ...,
        basic: str = ...,
        child_id: str = ...,
        child: __stubgen_tests_streams_test__window_ChildWindow = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.uid: str
        self.scalar_feature__600__: str
        self.scalar_feature__1200__: str
        self.scalar_feature: Windowed[str]
        self.scalar_feature_2__600__: str
        self.scalar_feature_2__1200__: str
        self.scalar_feature_2: Windowed[str]
        self.basic: str
        self.child_id: str
        self.child: __stubgen_tests_streams_test__window_ChildWindow
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class StreamFeaturesWindowProtocol(Protocol):
    uid: str
    scalar_feature: Windowed[str]
    scalar_feature_2: Windowed[str]
    basic: str
    child_id: str
    child: __stubgen_tests_streams_test__window_ChildWindow

class SQLExecuteFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def str_feature(self) -> Type[str]: ...

    @property
    def int_feature(self) -> Type[int]: ...

    @property
    def bool_feature(self) -> Type[bool]: ...

    @property
    def enum_feature(self) -> Type[__stubgen_tests_sql_test__execute_EnumFeature]: ...

    @property
    def float_feature(self) -> Type[float]: ...

    @property
    def datetime_feature(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def list_feature(self) -> Type[list[str]]: ...

    @property
    def nested_features_id(self) -> Type[int]: ...

    @property
    def nested_features(self) -> Type[__stubgen_tests_sql_test__execute_SQLExecuteNestedFeatures]: ...

class SQLExecuteFeatures(Features, metaclass=SQLExecuteFeaturesMetaclass):
    def __init__(
        self,
        id: int = ...,
        ts: __stubgen_datetime_datetime = ...,
        str_feature: str = ...,
        int_feature: int = ...,
        bool_feature: bool = ...,
        enum_feature: __stubgen_tests_sql_test__execute_EnumFeature = ...,
        float_feature: float = ...,
        datetime_feature: __stubgen_datetime_datetime = ...,
        list_feature: list[str] = ...,
        nested_features_id: int = ...,
        nested_features: __stubgen_tests_sql_test__execute_SQLExecuteNestedFeatures = ...,
    ):
        self.id: int
        self.ts: __stubgen_datetime_datetime
        self.str_feature: str
        self.int_feature: int
        self.bool_feature: bool
        self.enum_feature: __stubgen_tests_sql_test__execute_EnumFeature
        self.float_feature: float
        self.datetime_feature: __stubgen_datetime_datetime
        self.list_feature: list[str]
        self.nested_features_id: int
        self.nested_features: __stubgen_tests_sql_test__execute_SQLExecuteNestedFeatures

class SQLExecuteFeaturesProtocol(Protocol):
    id: int
    ts: __stubgen_datetime_datetime
    str_feature: str
    int_feature: int
    bool_feature: bool
    enum_feature: __stubgen_tests_sql_test__execute_EnumFeature
    float_feature: float
    datetime_feature: __stubgen_datetime_datetime
    list_feature: list[str]
    nested_features_id: int
    nested_features: __stubgen_tests_sql_test__execute_SQLExecuteNestedFeatures

class HelloMetaclass(FeaturesMeta):
    @property
    def a(self) -> Type[str]: ...

    @property
    def b(self) -> Type[int]: ...

    @property
    def c(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def d(self) -> Type[__stubgen_tests_serialization_test__codec_Color]: ...

    @property
    def e(self) -> Type[__stubgen_datetime_date]: ...

    @property
    def f(self) -> Type[__stubgen_tests_serialization_test__codec_MyDataclass]: ...

    @property
    def y(self) -> Type[set[int]]: ...

    @property
    def z(self) -> Type[list[str]]: ...

    @property
    def fancy(self) -> Type[__stubgen_tests_serialization_test__codec_CustomClass]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class Hello(Features, metaclass=HelloMetaclass):
    def __init__(
        self,
        a: str = ...,
        b: int = ...,
        c: __stubgen_datetime_datetime = ...,
        d: __stubgen_tests_serialization_test__codec_Color = ...,
        e: __stubgen_datetime_date = ...,
        f: __stubgen_tests_serialization_test__codec_MyDataclass = ...,
        y: set[int] = ...,
        z: list[str] = ...,
        fancy: __stubgen_tests_serialization_test__codec_CustomClass = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.a: str
        self.b: int
        self.c: __stubgen_datetime_datetime
        self.d: __stubgen_tests_serialization_test__codec_Color
        self.e: __stubgen_datetime_date
        self.f: __stubgen_tests_serialization_test__codec_MyDataclass
        self.y: set[int]
        self.z: list[str]
        self.fancy: __stubgen_tests_serialization_test__codec_CustomClass
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class HelloProtocol(Protocol):
    a: str
    b: int
    c: __stubgen_datetime_datetime
    d: __stubgen_tests_serialization_test__codec_Color
    e: __stubgen_datetime_date
    f: __stubgen_tests_serialization_test__codec_MyDataclass
    y: set[int]
    z: list[str]
    fancy: __stubgen_tests_serialization_test__codec_CustomClass

class FSWithStalenessMetaclass(FeaturesMeta):
    @property
    def parent_id(self) -> Type[str]: ...

    @property
    def scalar(self) -> Type[int]: ...

    @property
    def windowed_int__600__(self) -> Type[int]: ...

    @property
    def windowed_int(self) -> Type[Windowed[int]]: ...

    @property
    def scalar_override(self) -> Type[int]: ...

    @property
    def windowed_int_override__600__(self) -> Type[int]: ...

    @property
    def windowed_int_override(self) -> Type[Windowed[int]]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class FSWithStaleness(Features, metaclass=FSWithStalenessMetaclass):
    def __init__(
        self,
        parent_id: str = ...,
        scalar: int = ...,
        windowed_int__600__: int = ...,
        windowed_int: Windowed[int] = ...,
        scalar_override: int = ...,
        windowed_int_override__600__: int = ...,
        windowed_int_override: Windowed[int] = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.parent_id: str
        self.scalar: int
        self.windowed_int__600__: int
        self.windowed_int: Windowed[int]
        self.scalar_override: int
        self.windowed_int_override__600__: int
        self.windowed_int_override: Windowed[int]
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class FSWithStalenessProtocol(Protocol):
    parent_id: str
    scalar: int
    windowed_int: Windowed[int]
    scalar_override: int
    windowed_int_override: Windowed[int]

class MappingFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def u_from(self) -> Type[str]: ...

    @property
    def u_to(self) -> Type[str]: ...

    @property
    def column_a(self) -> Type[str]: ...

    @property
    def swap_b(self) -> Type[str]: ...

    @property
    def swap_c(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class MappingFeatures(Features, metaclass=MappingFeaturesMetaclass):
    def __init__(
        self,
        id: int = ...,
        u_from: str = ...,
        u_to: str = ...,
        column_a: str = ...,
        swap_b: str = ...,
        swap_c: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: int
        self.u_from: str
        self.u_to: str
        self.column_a: str
        self.swap_b: str
        self.swap_c: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class MappingFeaturesProtocol(Protocol):
    id: int
    u_from: str
    u_to: str
    column_a: str
    swap_b: str
    swap_c: str

class CommentBaseOwnerMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def empty(self) -> Type[str]: ...

    @property
    def email(self) -> Type[str]: ...

    @property
    def email_commas(self) -> Type[str]: ...

    @property
    def email_single(self) -> Type[str]: ...

    @property
    def email_all_kinds(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class CommentBaseOwner(Features, metaclass=CommentBaseOwnerMetaclass):
    def __init__(
        self,
        id: str = ...,
        empty: str = ...,
        email: str = ...,
        email_commas: str = ...,
        email_single: str = ...,
        email_all_kinds: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.empty: str
        self.email: str
        self.email_commas: str
        self.email_single: str
        self.email_all_kinds: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class CommentBaseOwnerProtocol(Protocol):
    id: str
    empty: str
    email: str
    email_commas: str
    email_single: str
    email_all_kinds: str

class WindowTypeMetaclass(FeaturesMeta):
    @property
    def uid(self) -> Type[str]: ...

    @property
    def str_window__600__(self) -> Type[str]: ...

    @property
    def str_window(self) -> Type[Windowed[str]]: ...

    @property
    def int_window__600__(self) -> Type[int]: ...

    @property
    def int_window(self) -> Type[Windowed[int]]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class WindowType(Features, metaclass=WindowTypeMetaclass):
    def __init__(
        self,
        uid: str = ...,
        str_window__600__: str = ...,
        str_window: Windowed[str] = ...,
        int_window__600__: int = ...,
        int_window: Windowed[int] = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.uid: str
        self.str_window__600__: str
        self.str_window: Windowed[str]
        self.int_window__600__: int
        self.int_window: Windowed[int]
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class WindowTypeProtocol(Protocol):
    uid: str
    str_window: Windowed[str]
    int_window: Windowed[int]

class PersonMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def name(self) -> Type[str]: ...

    @property
    def best_foot_id(self) -> Type[int]: ...

    @property
    def best_foot(self) -> Type[__stubgen_tests_client_test__expand__features_Foot]: ...

    @property
    def feet(self) -> Type[DataFrame]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class Person(Features, metaclass=PersonMetaclass):
    def __init__(
        self,
        id: int = ...,
        name: str = ...,
        best_foot_id: int = ...,
        best_foot: __stubgen_tests_client_test__expand__features_Foot = ...,
        feet: DataFrame = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: int
        self.name: str
        self.best_foot_id: int
        self.best_foot: __stubgen_tests_client_test__expand__features_Foot
        self.feet: DataFrame
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class PersonProtocol(Protocol):
    id: int
    name: str
    best_foot_id: int
    best_foot: __stubgen_tests_client_test__expand__features_Foot
    feet: DataFrame

class HomeFeaturesChainedFeatureTimeMetaclass(FeaturesMeta):
    @property
    def home_id(self) -> Type[str]: ...

    @property
    def address(self) -> Type[str]: ...

    @property
    def price(self) -> Type[int]: ...

    @property
    def sq_ft(self) -> Type[int]: ...

    @property
    def homeowner(self) -> Type[__stubgen_tests_features_test__chained__feature__time_Homeowner]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class HomeFeaturesChainedFeatureTime(Features, metaclass=HomeFeaturesChainedFeatureTimeMetaclass):
    def __init__(
        self,
        home_id: str = ...,
        address: str = ...,
        price: int = ...,
        sq_ft: int = ...,
        homeowner: __stubgen_tests_features_test__chained__feature__time_Homeowner = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.home_id: str
        self.address: str
        self.price: int
        self.sq_ft: int
        self.homeowner: __stubgen_tests_features_test__chained__feature__time_Homeowner
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class HomeFeaturesChainedFeatureTimeProtocol(Protocol):
    home_id: str
    address: str
    price: int
    sq_ft: int
    homeowner: __stubgen_tests_features_test__chained__feature__time_Homeowner

class ChildWindowMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def childfeat__600__(self) -> Type[str]: ...

    @property
    def childfeat__1200__(self) -> Type[str]: ...

    @property
    def childfeat(self) -> Type[Windowed[str]]: ...

    @property
    def grand(self) -> Type[__stubgen_tests_streams_test__window_GrandchildWindow]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class ChildWindow(Features, metaclass=ChildWindowMetaclass):
    def __init__(
        self,
        id: str = ...,
        childfeat__600__: str = ...,
        childfeat__1200__: str = ...,
        childfeat: Windowed[str] = ...,
        grand: __stubgen_tests_streams_test__window_GrandchildWindow = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.childfeat__600__: str
        self.childfeat__1200__: str
        self.childfeat: Windowed[str]
        self.grand: __stubgen_tests_streams_test__window_GrandchildWindow
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class ChildWindowProtocol(Protocol):
    id: str
    childfeat: Windowed[str]
    grand: __stubgen_tests_streams_test__window_GrandchildWindow

class TransactionMetaclass(FeaturesMeta):
    @property
    def user_id(self) -> Type[str]: ...

    @property
    def amount(self) -> Type[int]: ...

    @property
    def source_account_id(self) -> Type[int]: ...

    @property
    def dest_account_id(self) -> Type[int]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class Transaction(Features, metaclass=TransactionMetaclass):
    def __init__(
        self,
        user_id: str = ...,
        amount: int = ...,
        source_account_id: int = ...,
        dest_account_id: int = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.user_id: str
        self.amount: int
        self.source_account_id: int
        self.dest_account_id: int
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class TransactionProtocol(Protocol):
    user_id: str
    amount: int
    source_account_id: int
    dest_account_id: int

class ToppingMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def value(self) -> Type[int]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def nicknames(self) -> Type[DataFrame]: ...

    @property
    def price(self) -> Type[__stubgen_tests_features_test__df_ToppingPrice]: ...

class Topping(Features, metaclass=ToppingMetaclass):
    def __init__(
        self,
        id: str = ...,
        value: int = ...,
        ts: __stubgen_datetime_datetime = ...,
        nicknames: DataFrame = ...,
        price: __stubgen_tests_features_test__df_ToppingPrice = ...,
    ):
        self.id: str
        self.value: int
        self.ts: __stubgen_datetime_datetime
        self.nicknames: DataFrame
        self.price: __stubgen_tests_features_test__df_ToppingPrice

class ToppingProtocol(Protocol):
    id: str
    value: int
    ts: __stubgen_datetime_datetime
    nicknames: DataFrame
    price: __stubgen_tests_features_test__df_ToppingPrice

class StoreFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def purchases__600__(self) -> Type[float]: ...

    @property
    def purchases__1200__(self) -> Type[float]: ...

    @property
    def purchases(self) -> Type[Windowed[float]]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class StoreFeatures(Features, metaclass=StoreFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        purchases__600__: float = ...,
        purchases__1200__: float = ...,
        purchases: Windowed[float] = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.purchases__600__: float
        self.purchases__1200__: float
        self.purchases: Windowed[float]
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class StoreFeaturesProtocol(Protocol):
    id: str
    purchases: Windowed[float]

class MypyUserFeaturesMetaclass(FeaturesMeta):
    @property
    def uid(self) -> Type[str]: ...

    @property
    def name(self) -> Type[str]: ...

    @property
    def bday(self) -> Type[str]: ...

    @property
    def age(self) -> Type[int]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class MypyUserFeatures(Features, metaclass=MypyUserFeaturesMetaclass):
    def __init__(
        self,
        uid: str = ...,
        name: str = ...,
        bday: str = ...,
        age: int = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.uid: str
        self.name: str
        self.bday: str
        self.age: int
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class MypyUserFeaturesProtocol(Protocol):
    uid: str
    name: str
    bday: str
    age: int

class MaxStalenessFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def name(self) -> Type[str]: ...

    @property
    def woohoo(self) -> Type[str]: ...

    @property
    def boop(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class MaxStalenessFeatures(Features, metaclass=MaxStalenessFeaturesMetaclass):
    def __init__(
        self,
        id: int = ...,
        name: str = ...,
        woohoo: str = ...,
        boop: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: int
        self.name: str
        self.woohoo: str
        self.boop: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class MaxStalenessFeaturesProtocol(Protocol):
    id: int
    name: str
    woohoo: str
    boop: str

class HomeFeaturesMetaclass(FeaturesMeta):
    @property
    def home_id(self) -> Type[str]: ...

    @property
    def address(self) -> Type[str]: ...

    @property
    def price(self) -> Type[int]: ...

    @property
    def sq_ft(self) -> Type[int]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class HomeFeatures(Features, metaclass=HomeFeaturesMetaclass):
    def __init__(
        self,
        home_id: str = ...,
        address: str = ...,
        price: int = ...,
        sq_ft: int = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.home_id: str
        self.address: str
        self.price: int
        self.sq_ft: int
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class HomeFeaturesProtocol(Protocol):
    home_id: str
    address: str
    price: int
    sq_ft: int

class GrandchildWindowMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def grandchildfeat__600__(self) -> Type[str]: ...

    @property
    def grandchildfeat__1200__(self) -> Type[str]: ...

    @property
    def grandchildfeat(self) -> Type[Windowed[str]]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class GrandchildWindow(Features, metaclass=GrandchildWindowMetaclass):
    def __init__(
        self,
        id: str = ...,
        grandchildfeat__600__: str = ...,
        grandchildfeat__1200__: str = ...,
        grandchildfeat: Windowed[str] = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.grandchildfeat__600__: str
        self.grandchildfeat__1200__: str
        self.grandchildfeat: Windowed[str]
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class GrandchildWindowProtocol(Protocol):
    id: str
    grandchildfeat: Windowed[str]

class FootMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def is_flat(self) -> Type[bool]: ...

    @property
    def person_id(self) -> Type[int]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def ankle(self) -> Type[__stubgen_tests_client_test__expand__features_Ankle]: ...

class Foot(Features, metaclass=FootMetaclass):
    def __init__(
        self,
        id: int = ...,
        is_flat: bool = ...,
        person_id: int = ...,
        ts: __stubgen_datetime_datetime = ...,
        ankle: __stubgen_tests_client_test__expand__features_Ankle = ...,
    ):
        self.id: int
        self.is_flat: bool
        self.person_id: int
        self.ts: __stubgen_datetime_datetime
        self.ankle: __stubgen_tests_client_test__expand__features_Ankle

class FootProtocol(Protocol):
    id: int
    is_flat: bool
    person_id: int
    ts: __stubgen_datetime_datetime
    ankle: __stubgen_tests_client_test__expand__features_Ankle

class ContinuousFeatureClassMetaclass(FeaturesMeta):
    @property
    def c_feat__600__(self) -> Type[int]: ...

    @property
    def c_feat(self) -> Type[Windowed[int]]: ...

    @property
    def t_feat__600__(self) -> Type[int]: ...

    @property
    def t_feat(self) -> Type[Windowed[int]]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class ContinuousFeatureClass(Features, metaclass=ContinuousFeatureClassMetaclass):
    def __init__(
        self,
        c_feat__600__: int = ...,
        c_feat: Windowed[int] = ...,
        t_feat__600__: int = ...,
        t_feat: Windowed[int] = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.c_feat__600__: int
        self.c_feat: Windowed[int]
        self.t_feat__600__: int
        self.t_feat: Windowed[int]
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class ContinuousFeatureClassProtocol(Protocol):
    c_feat: Windowed[int]
    t_feat: Windowed[int]

class UserProfileMetaclass(FeaturesMeta):
    @property
    def user_id(self) -> Type[str]: ...

    @property
    def profile_id(self) -> Type[str]: ...

    @property
    def address(self) -> Type[int]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class UserProfile(Features, metaclass=UserProfileMetaclass):
    def __init__(
        self,
        user_id: str = ...,
        profile_id: str = ...,
        address: int = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.user_id: str
        self.profile_id: str
        self.address: int
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class UserProfileProtocol(Protocol):
    user_id: str
    profile_id: str
    address: int

class UserMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def fav_color(self) -> Type[__stubgen_tests_client_test__client__serialization_Color]: ...

    @property
    def birthday(self) -> Type[__stubgen_datetime_date]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class User(Features, metaclass=UserMetaclass):
    def __init__(
        self,
        id: int = ...,
        fav_color: __stubgen_tests_client_test__client__serialization_Color = ...,
        birthday: __stubgen_datetime_date = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: int
        self.fav_color: __stubgen_tests_client_test__client__serialization_Color
        self.birthday: __stubgen_datetime_date
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class UserProtocol(Protocol):
    id: int
    fav_color: __stubgen_tests_client_test__client__serialization_Color
    birthday: __stubgen_datetime_date

class ToppingPriceMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def topping_id(self) -> Type[str]: ...

    @property
    def wow(self) -> Type[str]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

class ToppingPrice(Features, metaclass=ToppingPriceMetaclass):
    def __init__(
        self,
        id: str = ...,
        topping_id: str = ...,
        wow: str = ...,
        ts: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.topping_id: str
        self.wow: str
        self.ts: __stubgen_datetime_datetime

class ToppingPriceProtocol(Protocol):
    id: str
    topping_id: str
    wow: str
    ts: __stubgen_datetime_datetime

class TagFeaturesMetaclass(FeaturesMeta):
    @property
    def empty(self) -> Type[str]: ...

    @property
    def one(self) -> Type[str]: ...

    @property
    def many(self) -> Type[str]: ...

    @property
    def ft(self) -> Type[__stubgen_datetime_datetime]: ...

class TagFeatures(Features, metaclass=TagFeaturesMetaclass):
    def __init__(
        self,
        empty: str = ...,
        one: str = ...,
        many: str = ...,
        ft: __stubgen_datetime_datetime = ...,
    ):
        self.empty: str
        self.one: str
        self.many: str
        self.ft: __stubgen_datetime_datetime

class TagFeaturesProtocol(Protocol):
    empty: str
    one: str
    many: str
    ft: __stubgen_datetime_datetime

class SQLFriendsWithRowMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def u_from(self) -> Type[str]: ...

    @property
    def u_to(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class SQLFriendsWithRow(Features, metaclass=SQLFriendsWithRowMetaclass):
    def __init__(
        self,
        id: int = ...,
        u_from: str = ...,
        u_to: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: int
        self.u_from: str
        self.u_to: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class SQLFriendsWithRowProtocol(Protocol):
    id: int
    u_from: str
    u_to: str

class ParentFSMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def children(self) -> Type[DataFrame]: ...

    @property
    def single_child(self) -> Type[__stubgen_tests_features_test__features_SingleChildFS]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

class ParentFS(Features, metaclass=ParentFSMetaclass):
    def __init__(
        self,
        id: str = ...,
        children: DataFrame = ...,
        single_child: __stubgen_tests_features_test__features_SingleChildFS = ...,
        ts: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.children: DataFrame
        self.single_child: __stubgen_tests_features_test__features_SingleChildFS
        self.ts: __stubgen_datetime_datetime

class ParentFSProtocol(Protocol):
    id: str
    children: DataFrame
    single_child: __stubgen_tests_features_test__features_SingleChildFS
    ts: __stubgen_datetime_datetime

class OwnerFeaturesMetaclass(FeaturesMeta):
    @property
    def plain(self) -> Type[str]: ...

    @property
    def cached(self) -> Type[str]: ...

    @property
    def andy(self) -> Type[str]: ...

    @property
    def ft(self) -> Type[__stubgen_datetime_datetime]: ...

class OwnerFeatures(Features, metaclass=OwnerFeaturesMetaclass):
    def __init__(
        self,
        plain: str = ...,
        cached: str = ...,
        andy: str = ...,
        ft: __stubgen_datetime_datetime = ...,
    ):
        self.plain: str
        self.cached: str
        self.andy: str
        self.ft: __stubgen_datetime_datetime

class OwnerFeaturesProtocol(Protocol):
    plain: str
    cached: str
    andy: str
    ft: __stubgen_datetime_datetime

class OtherPrimaryKeyFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def nah_really_this_is_id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class OtherPrimaryKeyFeatures(Features, metaclass=OtherPrimaryKeyFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        nah_really_this_is_id: str = ...,
        other: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.nah_really_this_is_id: str
        self.other: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class OtherPrimaryKeyFeaturesProtocol(Protocol):
    id: str
    nah_really_this_is_id: str
    other: str

class FunFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def nope(self) -> Type[str]: ...

    @property
    def single_parent(self) -> Type[__stubgen_tests_features_test__iter_NoFunFeatures]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

class FunFeatures(Features, metaclass=FunFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        nope: str = ...,
        single_parent: __stubgen_tests_features_test__iter_NoFunFeatures = ...,
        ts: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.nope: str
        self.single_parent: __stubgen_tests_features_test__iter_NoFunFeatures
        self.ts: __stubgen_datetime_datetime

class FunFeaturesProtocol(Protocol):
    id: str
    nope: str
    single_parent: __stubgen_tests_features_test__iter_NoFunFeatures
    ts: __stubgen_datetime_datetime

class ETLFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def name(self) -> Type[str]: ...

    @property
    def woohoo(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class ETLFeatures(Features, metaclass=ETLFeaturesMetaclass):
    def __init__(
        self,
        id: int = ...,
        name: str = ...,
        woohoo: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: int
        self.name: str
        self.woohoo: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class ETLFeaturesProtocol(Protocol):
    id: int
    name: str
    woohoo: str

class ColNameMappingFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def foo(self) -> Type[str]: ...

    @property
    def bar(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class ColNameMappingFeatures(Features, metaclass=ColNameMappingFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        foo: str = ...,
        bar: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.foo: str
        self.bar: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class ColNameMappingFeaturesProtocol(Protocol):
    id: str
    foo: str
    bar: str

class ChildFSMetaclass(FeaturesMeta):
    @property
    def parent_id(self) -> Type[str]: ...

    @property
    def parents(self) -> Type[DataFrame]: ...

    @property
    def single_parent(self) -> Type[__stubgen_tests_features_test__features_SingleParentFS]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class ChildFS(Features, metaclass=ChildFSMetaclass):
    def __init__(
        self,
        parent_id: str = ...,
        parents: DataFrame = ...,
        single_parent: __stubgen_tests_features_test__features_SingleParentFS = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.parent_id: str
        self.parents: DataFrame
        self.single_parent: __stubgen_tests_features_test__features_SingleParentFS
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class ChildFSProtocol(Protocol):
    parent_id: str
    parents: DataFrame
    single_parent: __stubgen_tests_features_test__features_SingleParentFS

class UnassignedIdFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class UnassignedIdFeatures(Features, metaclass=UnassignedIdFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.other: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class UnassignedIdFeaturesProtocol(Protocol):
    id: str
    other: str

class UnassignedDecoratedIdFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class UnassignedDecoratedIdFeatures(Features, metaclass=UnassignedDecoratedIdFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.other: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class UnassignedDecoratedIdFeaturesProtocol(Protocol):
    id: str
    other: str

class TheArtistFormerlyKnownAsPrinceMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def favorite_color(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class TheArtistFormerlyKnownAsPrince(Features, metaclass=TheArtistFormerlyKnownAsPrinceMetaclass):
    def __init__(
        self,
        id: str = ...,
        favorite_color: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.favorite_color: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class TheArtistFormerlyKnownAsPrinceProtocol(Protocol):
    id: str
    favorite_color: str

class SingleParentFSMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def children(self) -> Type[DataFrame]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class SingleParentFS(Features, metaclass=SingleParentFSMetaclass):
    def __init__(
        self,
        id: str = ...,
        children: DataFrame = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.children: DataFrame
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class SingleParentFSProtocol(Protocol):
    id: str
    children: DataFrame

class SingleChildFSMetaclass(FeaturesMeta):
    @property
    def parent_id(self) -> Type[str]: ...

    @property
    def parent(self) -> Type[DataFrame]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class SingleChildFS(Features, metaclass=SingleChildFSMetaclass):
    def __init__(
        self,
        parent_id: str = ...,
        parent: DataFrame = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.parent_id: str
        self.parent: DataFrame
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class SingleChildFSProtocol(Protocol):
    parent_id: str
    parent: DataFrame

class SQLUserFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def name(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class SQLUserFeatures(Features, metaclass=SQLUserFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        name: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.name: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class SQLUserFeaturesProtocol(Protocol):
    id: str
    name: str

class SQLExecuteNestedFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def nested_int(self) -> Type[int]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class SQLExecuteNestedFeatures(Features, metaclass=SQLExecuteNestedFeaturesMetaclass):
    def __init__(
        self,
        id: int = ...,
        nested_int: int = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: int
        self.nested_int: int
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class SQLExecuteNestedFeaturesProtocol(Protocol):
    id: int
    nested_int: int

class PrimaryCls2Metaclass(FeaturesMeta):
    @property
    def myid(self) -> Type[str]: ...

    @property
    def id(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class PrimaryCls2(Features, metaclass=PrimaryCls2Metaclass):
    def __init__(
        self,
        myid: str = ...,
        id: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.myid: str
        self.id: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class PrimaryCls2Protocol(Protocol):
    myid: str
    id: str

class NotIdIsIdFeaturesMetaclass(FeaturesMeta):
    @property
    def not_id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class NotIdIsIdFeatures(Features, metaclass=NotIdIsIdFeaturesMetaclass):
    def __init__(
        self,
        not_id: str = ...,
        other: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.not_id: str
        self.other: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class NotIdIsIdFeaturesProtocol(Protocol):
    not_id: str
    other: str

class IdIsNotIdFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class IdIsNotIdFeatures(Features, metaclass=IdIsNotIdFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.other: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class IdIsNotIdFeaturesProtocol(Protocol):
    id: str
    other: str

class HomeownerMetaclass(FeaturesMeta):
    @property
    def fullname(self) -> Type[str]: ...

    @property
    def home_id(self) -> Type[str]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

class Homeowner(Features, metaclass=HomeownerMetaclass):
    def __init__(
        self,
        fullname: str = ...,
        home_id: str = ...,
        ts: __stubgen_datetime_datetime = ...,
    ):
        self.fullname: str
        self.home_id: str
        self.ts: __stubgen_datetime_datetime

class HomeownerProtocol(Protocol):
    fullname: str
    home_id: str
    ts: __stubgen_datetime_datetime

class FTCls2Metaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def timestamp(self) -> Type[__stubgen_datetime_datetime]: ...

class FTCls2(Features, metaclass=FTCls2Metaclass):
    def __init__(
        self,
        id: str = ...,
        ts: __stubgen_datetime_datetime = ...,
        timestamp: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.ts: __stubgen_datetime_datetime
        self.timestamp: __stubgen_datetime_datetime

class FTCls2Protocol(Protocol):
    id: str
    ts: __stubgen_datetime_datetime
    timestamp: __stubgen_datetime_datetime

class ExplicitIdFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class ExplicitIdFeatures(Features, metaclass=ExplicitIdFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.other: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class ExplicitIdFeaturesProtocol(Protocol):
    id: str
    other: str

class ExampleFraudUserMetaclass(FeaturesMeta):
    @property
    def uid(self) -> Type[str]: ...

    @property
    def org(self) -> Type[__stubgen_tests_features_test__chained__has__one_ExampleFraudOrg]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class ExampleFraudUser(Features, metaclass=ExampleFraudUserMetaclass):
    def __init__(
        self,
        uid: str = ...,
        org: __stubgen_tests_features_test__chained__has__one_ExampleFraudOrg = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.uid: str
        self.org: __stubgen_tests_features_test__chained__has__one_ExampleFraudOrg
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class ExampleFraudUserProtocol(Protocol):
    uid: str
    org: __stubgen_tests_features_test__chained__has__one_ExampleFraudOrg

class ExampleFraudProfileMetaclass(FeaturesMeta):
    @property
    def uid(self) -> Type[str]: ...

    @property
    def user(self) -> Type[__stubgen_tests_features_test__chained__has__one_ExampleFraudUser]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class ExampleFraudProfile(Features, metaclass=ExampleFraudProfileMetaclass):
    def __init__(
        self,
        uid: str = ...,
        user: __stubgen_tests_features_test__chained__has__one_ExampleFraudUser = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.uid: str
        self.user: __stubgen_tests_features_test__chained__has__one_ExampleFraudUser
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class ExampleFraudProfileProtocol(Protocol):
    uid: str
    user: __stubgen_tests_features_test__chained__has__one_ExampleFraudUser

class ExampleFraudOrgMetaclass(FeaturesMeta):
    @property
    def uid(self) -> Type[str]: ...

    @property
    def org_name(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class ExampleFraudOrg(Features, metaclass=ExampleFraudOrgMetaclass):
    def __init__(
        self,
        uid: str = ...,
        org_name: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.uid: str
        self.org_name: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class ExampleFraudOrgProtocol(Protocol):
    uid: str
    org_name: str

class CustomNameClassMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class CustomNameClass(Features, metaclass=CustomNameClassMetaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.other: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class CustomNameClassProtocol(Protocol):
    id: str
    other: str

class BogusIdFeature2Metaclass(FeaturesMeta):
    @property
    def id(self) -> Type[DataFrame]: ...

    @property
    def other(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class BogusIdFeature2(Features, metaclass=BogusIdFeature2Metaclass):
    def __init__(
        self,
        id: DataFrame = ...,
        other: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: DataFrame
        self.other: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class BogusIdFeature2Protocol(Protocol):
    id: DataFrame
    other: str

class BogusIdFeature1Metaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class BogusIdFeature1(Features, metaclass=BogusIdFeature1Metaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.other: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class BogusIdFeature1Protocol(Protocol):
    id: str
    other: str

class AnkleMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def foot_id(self) -> Type[int]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class Ankle(Features, metaclass=AnkleMetaclass):
    def __init__(
        self,
        id: int = ...,
        foot_id: int = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: int
        self.foot_id: int
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class AnkleProtocol(Protocol):
    id: int
    foot_id: int

class AcctMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def balance(self) -> Type[float]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class Acct(Features, metaclass=AcctMetaclass):
    def __init__(
        self,
        id: str = ...,
        balance: float = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.balance: float
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class AcctProtocol(Protocol):
    id: str
    balance: float

class AccountMetaclass(FeaturesMeta):
    @property
    def account_id(self) -> Type[str]: ...

    @property
    def balance(self) -> Type[int]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class Account(Features, metaclass=AccountMetaclass):
    def __init__(
        self,
        account_id: str = ...,
        balance: int = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.account_id: str
        self.balance: int
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class AccountProtocol(Protocol):
    account_id: str
    balance: int

class UserFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class UserFeatures(Features, metaclass=UserFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class UserFeaturesProtocol(Protocol):
    id: str

class StreamFeaturesMetaclass(FeaturesMeta):
    @property
    def scalar_feature(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class StreamFeatures(Features, metaclass=StreamFeaturesMetaclass):
    def __init__(
        self,
        scalar_feature: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.scalar_feature: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class StreamFeaturesProtocol(Protocol):
    scalar_feature: str

class PrimaryClsMetaclass(FeaturesMeta):
    @property
    def myid(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class PrimaryCls(Features, metaclass=PrimaryClsMetaclass):
    def __init__(
        self,
        myid: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.myid: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class PrimaryClsProtocol(Protocol):
    myid: str

class NoFunFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class NoFunFeatures(Features, metaclass=NoFunFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class NoFunFeaturesProtocol(Protocol):
    id: str

class NicknameMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class Nickname(Features, metaclass=NicknameMetaclass):
    def __init__(
        self,
        id: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class NicknameProtocol(Protocol):
    id: str

class MaxStalenessFeatures2Metaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class MaxStalenessFeatures2(Features, metaclass=MaxStalenessFeatures2Metaclass):
    def __init__(
        self,
        id: int = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: int
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class MaxStalenessFeatures2Protocol(Protocol):
    id: int

class LibraryFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class LibraryFeatures(Features, metaclass=LibraryFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class LibraryFeaturesProtocol(Protocol):
    id: str

class FTCls1Metaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

class FTCls1(Features, metaclass=FTCls1Metaclass):
    def __init__(
        self,
        id: str = ...,
        ts: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.ts: __stubgen_datetime_datetime

class FTCls1Protocol(Protocol):
    id: str
    ts: __stubgen_datetime_datetime

class FTClsMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def timestamp(self) -> Type[__stubgen_datetime_datetime]: ...

class FTCls(Features, metaclass=FTClsMetaclass):
    def __init__(
        self,
        id: str = ...,
        timestamp: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.timestamp: __stubgen_datetime_datetime

class FTClsProtocol(Protocol):
    id: str
    timestamp: __stubgen_datetime_datetime

class BogusIdFeature3Metaclass(FeaturesMeta):
    @property
    def id(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def other(self) -> Type[str]: ...

class BogusIdFeature3(Features, metaclass=BogusIdFeature3Metaclass):
    def __init__(
        self,
        id: __stubgen_datetime_datetime = ...,
        other: str = ...,
    ):
        self.id: __stubgen_datetime_datetime
        self.other: str

class BogusIdFeature3Protocol(Protocol):
    id: __stubgen_datetime_datetime
    other: str

class FeaturesClassWithoutTimestampMetaclass(FeaturesMeta):
    @property
    def __chalk_observed_at__(self) -> Type[__stubgen_datetime_datetime]: ...

class FeaturesClassWithoutTimestamp(Features, metaclass=FeaturesClassWithoutTimestampMetaclass):
    def __init__(
        self,
        __chalk_observed_at__: __stubgen_datetime_datetime = ...,
    ):
        self.__chalk_observed_at__: __stubgen_datetime_datetime

class FeaturesClassWithoutTimestampProtocol(Protocol):
    ...

class FeaturesClassWithNamedTsMetaclass(FeaturesMeta):
    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

class FeaturesClassWithNamedTs(Features, metaclass=FeaturesClassWithNamedTsMetaclass):
    def __init__(
        self,
        ts: __stubgen_datetime_datetime = ...,
    ):
        self.ts: __stubgen_datetime_datetime

class FeaturesClassWithNamedTsProtocol(Protocol):
    ts: __stubgen_datetime_datetime

class FeaturesClassWithCustomTsNameMetaclass(FeaturesMeta):
    @property
    def ts_custom_name(self) -> Type[__stubgen_datetime_datetime]: ...

class FeaturesClassWithCustomTsName(Features, metaclass=FeaturesClassWithCustomTsNameMetaclass):
    def __init__(
        self,
        ts_custom_name: __stubgen_datetime_datetime = ...,
    ):
        self.ts_custom_name: __stubgen_datetime_datetime

class FeaturesClassWithCustomTsNameProtocol(Protocol):
    ts_custom_name: __stubgen_datetime_datetime

@overload
def features(item: Type[TacoProtocol]) -> Type[Taco]: ...

@overload
def features(item: Type[WowFSProtocol]) -> Type[WowFS]: ...

@overload
def features(item: Type[StreamFeaturesWindowProtocol]) -> Type[StreamFeaturesWindow]: ...

@overload
def features(item: Type[SQLExecuteFeaturesProtocol]) -> Type[SQLExecuteFeatures]: ...

@overload
def features(item: Type[HelloProtocol]) -> Type[Hello]: ...

@overload
def features(item: Type[FSWithStalenessProtocol]) -> Type[FSWithStaleness]: ...

@overload
def features(item: Type[MappingFeaturesProtocol]) -> Type[MappingFeatures]: ...

@overload
def features(item: Type[CommentBaseOwnerProtocol]) -> Type[CommentBaseOwner]: ...

@overload
def features(item: Type[WindowTypeProtocol]) -> Type[WindowType]: ...

@overload
def features(item: Type[PersonProtocol]) -> Type[Person]: ...

@overload
def features(item: Type[HomeFeaturesChainedFeatureTimeProtocol]) -> Type[HomeFeaturesChainedFeatureTime]: ...

@overload
def features(item: Type[ChildWindowProtocol]) -> Type[ChildWindow]: ...

@overload
def features(item: Type[TransactionProtocol]) -> Type[Transaction]: ...

@overload
def features(item: Type[ToppingProtocol]) -> Type[Topping]: ...

@overload
def features(item: Type[StoreFeaturesProtocol]) -> Type[StoreFeatures]: ...

@overload
def features(item: Type[MypyUserFeaturesProtocol]) -> Type[MypyUserFeatures]: ...

@overload
def features(item: Type[MaxStalenessFeaturesProtocol]) -> Type[MaxStalenessFeatures]: ...

@overload
def features(item: Type[HomeFeaturesProtocol]) -> Type[HomeFeatures]: ...

@overload
def features(item: Type[GrandchildWindowProtocol]) -> Type[GrandchildWindow]: ...

@overload
def features(item: Type[FootProtocol]) -> Type[Foot]: ...

@overload
def features(item: Type[ContinuousFeatureClassProtocol]) -> Type[ContinuousFeatureClass]: ...

@overload
def features(item: Type[UserProfileProtocol]) -> Type[UserProfile]: ...

@overload
def features(item: Type[UserProtocol]) -> Type[User]: ...

@overload
def features(item: Type[ToppingPriceProtocol]) -> Type[ToppingPrice]: ...

@overload
def features(item: Type[TagFeaturesProtocol]) -> Type[TagFeatures]: ...

@overload
def features(item: Type[SQLFriendsWithRowProtocol]) -> Type[SQLFriendsWithRow]: ...

@overload
def features(item: Type[ParentFSProtocol]) -> Type[ParentFS]: ...

@overload
def features(item: Type[OwnerFeaturesProtocol]) -> Type[OwnerFeatures]: ...

@overload
def features(item: Type[OtherPrimaryKeyFeaturesProtocol]) -> Type[OtherPrimaryKeyFeatures]: ...

@overload
def features(item: Type[FunFeaturesProtocol]) -> Type[FunFeatures]: ...

@overload
def features(item: Type[ETLFeaturesProtocol]) -> Type[ETLFeatures]: ...

@overload
def features(item: Type[ColNameMappingFeaturesProtocol]) -> Type[ColNameMappingFeatures]: ...

@overload
def features(item: Type[ChildFSProtocol]) -> Type[ChildFS]: ...

@overload
def features(item: Type[UnassignedIdFeaturesProtocol]) -> Type[UnassignedIdFeatures]: ...

@overload
def features(item: Type[UnassignedDecoratedIdFeaturesProtocol]) -> Type[UnassignedDecoratedIdFeatures]: ...

@overload
def features(item: Type[TheArtistFormerlyKnownAsPrinceProtocol]) -> Type[TheArtistFormerlyKnownAsPrince]: ...

@overload
def features(item: Type[SingleParentFSProtocol]) -> Type[SingleParentFS]: ...

@overload
def features(item: Type[SingleChildFSProtocol]) -> Type[SingleChildFS]: ...

@overload
def features(item: Type[SQLUserFeaturesProtocol]) -> Type[SQLUserFeatures]: ...

@overload
def features(item: Type[SQLExecuteNestedFeaturesProtocol]) -> Type[SQLExecuteNestedFeatures]: ...

@overload
def features(item: Type[PrimaryCls2Protocol]) -> Type[PrimaryCls2]: ...

@overload
def features(item: Type[NotIdIsIdFeaturesProtocol]) -> Type[NotIdIsIdFeatures]: ...

@overload
def features(item: Type[IdIsNotIdFeaturesProtocol]) -> Type[IdIsNotIdFeatures]: ...

@overload
def features(item: Type[HomeownerProtocol]) -> Type[Homeowner]: ...

@overload
def features(item: Type[FTCls2Protocol]) -> Type[FTCls2]: ...

@overload
def features(item: Type[ExplicitIdFeaturesProtocol]) -> Type[ExplicitIdFeatures]: ...

@overload
def features(item: Type[ExampleFraudUserProtocol]) -> Type[ExampleFraudUser]: ...

@overload
def features(item: Type[ExampleFraudProfileProtocol]) -> Type[ExampleFraudProfile]: ...

@overload
def features(item: Type[ExampleFraudOrgProtocol]) -> Type[ExampleFraudOrg]: ...

@overload
def features(item: Type[CustomNameClassProtocol]) -> Type[CustomNameClass]: ...

@overload
def features(item: Type[BogusIdFeature2Protocol]) -> Type[BogusIdFeature2]: ...

@overload
def features(item: Type[BogusIdFeature1Protocol]) -> Type[BogusIdFeature1]: ...

@overload
def features(item: Type[AnkleProtocol]) -> Type[Ankle]: ...

@overload
def features(item: Type[AcctProtocol]) -> Type[Acct]: ...

@overload
def features(item: Type[AccountProtocol]) -> Type[Account]: ...

@overload
def features(item: Type[UserFeaturesProtocol]) -> Type[UserFeatures]: ...

@overload
def features(item: Type[StreamFeaturesProtocol]) -> Type[StreamFeatures]: ...

@overload
def features(item: Type[PrimaryClsProtocol]) -> Type[PrimaryCls]: ...

@overload
def features(item: Type[NoFunFeaturesProtocol]) -> Type[NoFunFeatures]: ...

@overload
def features(item: Type[NicknameProtocol]) -> Type[Nickname]: ...

@overload
def features(item: Type[MaxStalenessFeatures2Protocol]) -> Type[MaxStalenessFeatures2]: ...

@overload
def features(item: Type[LibraryFeaturesProtocol]) -> Type[LibraryFeatures]: ...

@overload
def features(item: Type[FTCls1Protocol]) -> Type[FTCls1]: ...

@overload
def features(item: Type[FTClsProtocol]) -> Type[FTCls]: ...

@overload
def features(item: Type[BogusIdFeature3Protocol]) -> Type[BogusIdFeature3]: ...

@overload
def features(item: Type[FeaturesClassWithoutTimestampProtocol]) -> Type[FeaturesClassWithoutTimestamp]: ...

@overload
def features(item: Type[FeaturesClassWithNamedTsProtocol]) -> Type[FeaturesClassWithNamedTs]: ...

@overload
def features(item: Type[FeaturesClassWithCustomTsNameProtocol]) -> Type[FeaturesClassWithCustomTsName]: ...

@overload
def features(
    *,
    owner: Optional[str] = ...,
    tags: Optional[Tags] = ...,
    max_staleness: Optional[Duration] = ...,
    etl_offline_to_online: Optional[bool] = ...,
) -> __stubgen__features_proto: ...

class __stubgen__features_proto(Protocol):
    @overload
    def __call__(self, item: Type[TacoProtocol]) -> Type[Taco]: ...

    @overload
    def __call__(self, item: Type[WowFSProtocol]) -> Type[WowFS]: ...

    @overload
    def __call__(self, item: Type[StreamFeaturesWindowProtocol]) -> Type[StreamFeaturesWindow]: ...

    @overload
    def __call__(self, item: Type[SQLExecuteFeaturesProtocol]) -> Type[SQLExecuteFeatures]: ...

    @overload
    def __call__(self, item: Type[HelloProtocol]) -> Type[Hello]: ...

    @overload
    def __call__(self, item: Type[FSWithStalenessProtocol]) -> Type[FSWithStaleness]: ...

    @overload
    def __call__(self, item: Type[MappingFeaturesProtocol]) -> Type[MappingFeatures]: ...

    @overload
    def __call__(self, item: Type[CommentBaseOwnerProtocol]) -> Type[CommentBaseOwner]: ...

    @overload
    def __call__(self, item: Type[WindowTypeProtocol]) -> Type[WindowType]: ...

    @overload
    def __call__(self, item: Type[PersonProtocol]) -> Type[Person]: ...

    @overload
    def __call__(self, item: Type[HomeFeaturesChainedFeatureTimeProtocol]) -> Type[HomeFeaturesChainedFeatureTime]: ...

    @overload
    def __call__(self, item: Type[ChildWindowProtocol]) -> Type[ChildWindow]: ...

    @overload
    def __call__(self, item: Type[TransactionProtocol]) -> Type[Transaction]: ...

    @overload
    def __call__(self, item: Type[ToppingProtocol]) -> Type[Topping]: ...

    @overload
    def __call__(self, item: Type[StoreFeaturesProtocol]) -> Type[StoreFeatures]: ...

    @overload
    def __call__(self, item: Type[MypyUserFeaturesProtocol]) -> Type[MypyUserFeatures]: ...

    @overload
    def __call__(self, item: Type[MaxStalenessFeaturesProtocol]) -> Type[MaxStalenessFeatures]: ...

    @overload
    def __call__(self, item: Type[HomeFeaturesProtocol]) -> Type[HomeFeatures]: ...

    @overload
    def __call__(self, item: Type[GrandchildWindowProtocol]) -> Type[GrandchildWindow]: ...

    @overload
    def __call__(self, item: Type[FootProtocol]) -> Type[Foot]: ...

    @overload
    def __call__(self, item: Type[ContinuousFeatureClassProtocol]) -> Type[ContinuousFeatureClass]: ...

    @overload
    def __call__(self, item: Type[UserProfileProtocol]) -> Type[UserProfile]: ...

    @overload
    def __call__(self, item: Type[UserProtocol]) -> Type[User]: ...

    @overload
    def __call__(self, item: Type[ToppingPriceProtocol]) -> Type[ToppingPrice]: ...

    @overload
    def __call__(self, item: Type[TagFeaturesProtocol]) -> Type[TagFeatures]: ...

    @overload
    def __call__(self, item: Type[SQLFriendsWithRowProtocol]) -> Type[SQLFriendsWithRow]: ...

    @overload
    def __call__(self, item: Type[ParentFSProtocol]) -> Type[ParentFS]: ...

    @overload
    def __call__(self, item: Type[OwnerFeaturesProtocol]) -> Type[OwnerFeatures]: ...

    @overload
    def __call__(self, item: Type[OtherPrimaryKeyFeaturesProtocol]) -> Type[OtherPrimaryKeyFeatures]: ...

    @overload
    def __call__(self, item: Type[FunFeaturesProtocol]) -> Type[FunFeatures]: ...

    @overload
    def __call__(self, item: Type[ETLFeaturesProtocol]) -> Type[ETLFeatures]: ...

    @overload
    def __call__(self, item: Type[ColNameMappingFeaturesProtocol]) -> Type[ColNameMappingFeatures]: ...

    @overload
    def __call__(self, item: Type[ChildFSProtocol]) -> Type[ChildFS]: ...

    @overload
    def __call__(self, item: Type[UnassignedIdFeaturesProtocol]) -> Type[UnassignedIdFeatures]: ...

    @overload
    def __call__(self, item: Type[UnassignedDecoratedIdFeaturesProtocol]) -> Type[UnassignedDecoratedIdFeatures]: ...

    @overload
    def __call__(self, item: Type[TheArtistFormerlyKnownAsPrinceProtocol]) -> Type[TheArtistFormerlyKnownAsPrince]: ...

    @overload
    def __call__(self, item: Type[SingleParentFSProtocol]) -> Type[SingleParentFS]: ...

    @overload
    def __call__(self, item: Type[SingleChildFSProtocol]) -> Type[SingleChildFS]: ...

    @overload
    def __call__(self, item: Type[SQLUserFeaturesProtocol]) -> Type[SQLUserFeatures]: ...

    @overload
    def __call__(self, item: Type[SQLExecuteNestedFeaturesProtocol]) -> Type[SQLExecuteNestedFeatures]: ...

    @overload
    def __call__(self, item: Type[PrimaryCls2Protocol]) -> Type[PrimaryCls2]: ...

    @overload
    def __call__(self, item: Type[NotIdIsIdFeaturesProtocol]) -> Type[NotIdIsIdFeatures]: ...

    @overload
    def __call__(self, item: Type[IdIsNotIdFeaturesProtocol]) -> Type[IdIsNotIdFeatures]: ...

    @overload
    def __call__(self, item: Type[HomeownerProtocol]) -> Type[Homeowner]: ...

    @overload
    def __call__(self, item: Type[FTCls2Protocol]) -> Type[FTCls2]: ...

    @overload
    def __call__(self, item: Type[ExplicitIdFeaturesProtocol]) -> Type[ExplicitIdFeatures]: ...

    @overload
    def __call__(self, item: Type[ExampleFraudUserProtocol]) -> Type[ExampleFraudUser]: ...

    @overload
    def __call__(self, item: Type[ExampleFraudProfileProtocol]) -> Type[ExampleFraudProfile]: ...

    @overload
    def __call__(self, item: Type[ExampleFraudOrgProtocol]) -> Type[ExampleFraudOrg]: ...

    @overload
    def __call__(self, item: Type[CustomNameClassProtocol]) -> Type[CustomNameClass]: ...

    @overload
    def __call__(self, item: Type[BogusIdFeature2Protocol]) -> Type[BogusIdFeature2]: ...

    @overload
    def __call__(self, item: Type[BogusIdFeature1Protocol]) -> Type[BogusIdFeature1]: ...

    @overload
    def __call__(self, item: Type[AnkleProtocol]) -> Type[Ankle]: ...

    @overload
    def __call__(self, item: Type[AcctProtocol]) -> Type[Acct]: ...

    @overload
    def __call__(self, item: Type[AccountProtocol]) -> Type[Account]: ...

    @overload
    def __call__(self, item: Type[UserFeaturesProtocol]) -> Type[UserFeatures]: ...

    @overload
    def __call__(self, item: Type[StreamFeaturesProtocol]) -> Type[StreamFeatures]: ...

    @overload
    def __call__(self, item: Type[PrimaryClsProtocol]) -> Type[PrimaryCls]: ...

    @overload
    def __call__(self, item: Type[NoFunFeaturesProtocol]) -> Type[NoFunFeatures]: ...

    @overload
    def __call__(self, item: Type[NicknameProtocol]) -> Type[Nickname]: ...

    @overload
    def __call__(self, item: Type[MaxStalenessFeatures2Protocol]) -> Type[MaxStalenessFeatures2]: ...

    @overload
    def __call__(self, item: Type[LibraryFeaturesProtocol]) -> Type[LibraryFeatures]: ...

    @overload
    def __call__(self, item: Type[FTCls1Protocol]) -> Type[FTCls1]: ...

    @overload
    def __call__(self, item: Type[FTClsProtocol]) -> Type[FTCls]: ...

    @overload
    def __call__(self, item: Type[BogusIdFeature3Protocol]) -> Type[BogusIdFeature3]: ...

    @overload
    def __call__(self, item: Type[FeaturesClassWithoutTimestampProtocol]) -> Type[FeaturesClassWithoutTimestamp]: ...

    @overload
    def __call__(self, item: Type[FeaturesClassWithNamedTsProtocol]) -> Type[FeaturesClassWithNamedTs]: ...

    @overload
    def __call__(self, item: Type[FeaturesClassWithCustomTsNameProtocol]) -> Type[FeaturesClassWithCustomTsName]: ...



import contextlib
import functools
import inspect
import os
import os.path
import warnings
from dataclasses import dataclass
from enum import Enum
from os import PathLike
from typing import Any, ClassVar, Dict, List, Mapping, Optional, Sequence, Set, Tuple, Type, TypeVar, Union

import pyarrow as pa
import sqlalchemy.sql.functions
from sqlalchemy import column, text
from sqlalchemy.engine import URL, Connection, Engine, create_engine
from sqlalchemy.orm import InstrumentedAttribute, Session
from sqlalchemy.sql.elements import Label

from chalk.features import Feature, Features, FeatureWrapper, unwrap_feature
from chalk.sql._internal.chalk_query import ChalkQuery
from chalk.sql._internal.incremental import IncrementalSettings
from chalk.sql._internal.string_chalk_query import StringChalkQuery
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.sql.protocols import BaseSQLSourceProtocol, ChalkQueryProtocol, StringChalkQueryProtocol, TableIngestProtocol
from chalk.utils.environment_parsing import env_var_bool

TTableIngestMixIn = TypeVar("TTableIngestMixIn", bound="TableIngestMixIn")
CHALK_QUERY_LOGGING = env_var_bool("CHALK_QUERY_LOGGING")


@dataclass
class TableIngestionPreferences:
    features: Type[Features]
    ignore_columns: Set[str]
    ignore_features: Set[str]
    require_columns: Set[str]
    require_features: Set[str]
    column_to_feature: Dict[str, str]
    cdc: Optional[Union[bool, IncrementalSettings]]


def _force_set_str(x: Optional[List[Any]]) -> Set[str]:
    return set() if x is None else set(map(str, x))


class TableIngestMixIn(TableIngestProtocol):
    ingested_tables: Dict[str, TableIngestionPreferences]

    def with_table(
        self: TTableIngestMixIn,
        *,
        name: str,
        features: Type[Union[Features, Any]],
        ignore_columns: Optional[List[str]] = None,
        ignore_features: Optional[List[Union[str, Any]]] = None,
        require_columns: Optional[List[str]] = None,
        require_features: Optional[List[Union[str, Any]]] = None,
        column_to_feature: Optional[Dict[str, Any]] = None,
        cdc: Optional[Union[bool, IncrementalSettings]] = None,
    ) -> TTableIngestMixIn:
        if name in self.ingested_tables:
            raise ValueError(f"The table {name} is ingested twice.")
        self.ingested_tables[name] = TableIngestionPreferences(
            features=features,
            ignore_columns=_force_set_str(ignore_columns),
            ignore_features=_force_set_str(ignore_features),
            require_columns=_force_set_str(require_columns),
            require_features=_force_set_str(require_features),
            column_to_feature={k: str(v) for k, v in (column_to_feature or {}).items()},
            cdc=cdc,
        )
        return self


class BaseSQLSource(BaseSQLSourceProtocol):
    registry: ClassVar[List["BaseSQLSource"]] = []

    def __init__(self):
        self._incremental_settings = None
        self.registry.append(self)

    def query_sql_file(
        self,
        path: Union[str, bytes, PathLike],
        fields: Optional[Mapping[str, Union[Feature, Any]]] = None,
        args: Optional[Mapping[str, str]] = None,
    ) -> StringChalkQueryProtocol:
        sql_string = None
        if os.path.isfile(path):
            with open(path) as f:
                sql_string = f.read()
        else:
            caller_filename = inspect.stack()[1].filename
            dir_path = os.path.dirname(os.path.realpath(caller_filename))
            if isinstance(path, bytes):
                path = path.decode("utf-8")
            relative_path = os.path.join(dir_path, path)
            if os.path.isfile(relative_path):
                with open(relative_path) as f:
                    sql_string = f.read()
        if sql_string is None:
            raise FileNotFoundError(f"No such file: '{str(path)}'")
        return self.query_string(
            query=sql_string,
            fields=fields,
            args=args,
        )

    def query_string(
        self,
        query: str,
        fields: Optional[Mapping[str, Union[Feature, Any]]] = None,
        args: Optional[Mapping[str, str]] = None,
    ) -> StringChalkQueryProtocol:
        fields = fields or {}
        fields = {f: unwrap_feature(v) if isinstance(v, FeatureWrapper) else v for (f, v) in fields.items()}
        return StringChalkQuery(source=self, query=query, fields=fields, params=args or {})

    def warm_up(self) -> None:
        with self.get_engine().connect() as cnx:
            cnx.execute(text("select 1"))

    def query(self, *entities: Any) -> ChalkQueryProtocol:
        targets: List[Label] = []
        features: Dict[str, Feature] = {}

        for e in entities:
            if isinstance(e, Features):
                _extract_features(e, e.__chalk_namespace__, targets, features)
            else:
                targets.append(e)

        return ChalkQuery(
            features=features,
            targets=targets,
            source=self,
        )

    def local_engine_url(self) -> URL:
        raise NotImplementedError

    def execute_query(self, finalized_query: FinalizedChalkQuery, connection: Optional[Any] = None) -> pa.Table:
        with (self.get_engine().connect() if connection is None else contextlib.nullcontext(connection)) as cnx:
            assert isinstance(cnx, Connection)
            res = cnx.execute(finalized_query.query, finalized_query.params)
            data: Dict[str, List[Any]] = {}
            desc = res.cursor.description
            result_columns: list[str] = [col[0] for col in desc]
            for v in result_columns:
                data[v] = []
            for row in res.all():
                for k, v in row._mapping.items():
                    if isinstance(v, Enum):
                        v = v.value
                    data[k].append(v)
            return pa.Table.from_pydict(data)

    def compile_query(
        self,
        finalized_query: FinalizedChalkQuery,
        paramstyle: Optional[str] = None,
    ) -> Tuple[str, Sequence[Any], Dict[str, Any]]:
        """Compile a query into a string and the bindparams"""
        dialect = self.local_engine_url().get_dialect()(paramstyle=paramstyle)
        query = finalized_query.query.params(finalized_query.params)
        compiled_query = query.compile(dialect=dialect)
        query_string = compiled_query.string
        return query_string, compiled_query.positiontup or [], compiled_query.params

    @functools.lru_cache(None)  # pyright: ignore[reportIncompatibleMethodOverride]
    def get_engine(self) -> Engine:
        return create_engine(
            url=self.local_engine_url(),
            echo=CHALK_QUERY_LOGGING,
            **self.engine_args(),
        )

    def raw_session(self) -> Session:
        warnings.warn(
            DeprecationWarning(
                (
                    "The method `raw_session()` is deprecated. Instead, please construct a session directly "
                    "from the underlying engine -- for example:"
                    "`from sqlalchemy.orm import Session; with Session(sql_source.get_engine()) as session: ...`"
                )
            )
        )
        return Session(self.get_engine())


def _extract_features(feature_set: Features, prefix: str, targets: List[Label], features: Dict[str, Feature]):
    for f in feature_set.features:
        assert f.attribute_name is not None
        try:
            feature_value = getattr(feature_set, f.attribute_name)
        except AttributeError:
            continue
        if f.is_has_many:
            raise ValueError(
                f"Feature '{prefix}.{f.name}' is a nested has-many feature, which is not supported when querying from SQL"
            )
        root_fqn = f"{prefix}.{f.name}"
        if isinstance(feature_value, Features):
            # Nested sub-features
            _extract_features(feature_value, root_fqn, targets, features)
        elif isinstance(feature_value, str):
            # Treat it as a column name
            features[root_fqn] = Feature.from_root_fqn(root_fqn)
            targets.append(column(feature_value).label(root_fqn))
        elif isinstance(feature_value, (sqlalchemy.sql.functions.GenericFunction, InstrumentedAttribute)):
            features[root_fqn] = Feature.from_root_fqn(root_fqn)
            targets.append(feature_value.label(root_fqn))
        else:
            raise TypeError(
                (
                    f"Feature '{root_fqn}' has an unsupported value for SQL queries. "
                    "All values must be a column, a column name string, a nested Features class, or an SQLAlchemy function"
                )
            )



from chalk.features import after_all, before_all
from chalk.features.hooks import Hook


@before_all
def simple_hook_init():
    return 4


@after_all
def simple_hook_teardown():
    return 5


@after_all(environment="hello")
def env_after_all():
    return 6


@before_all(environment=["hello", "itsme"])
def env_before_all():
    return 5


class Client:
    init = False


client = Client()
client2 = Client()


@before_all(environment="my-special-env")
def setup_env_1():
    client.init = True


@before_all(environment="my-special-env-2")
def setup_env_2():
    client2.init = True


def test_hooks_callable():
    assert simple_hook_init() == 4
    assert simple_hook_init.environment is None

    assert simple_hook_teardown() == 5
    assert simple_hook_teardown.environment is None


def test_hook_environments():
    assert env_before_all() == 5
    assert env_before_all.environment == ["hello", "itsme"]

    assert env_after_all() == 6
    assert env_after_all.environment == ["hello"]


def test_run_all():
    assert not client.init
    assert not client2.init
    Hook.run_all_before_all("my-special-env")
    assert client.init
    assert not client2.init



import concurrent.futures
import logging
from typing import Any, List, cast

import pytest

from chalk.utils.log_with_context import add_logging_context, get_logger


@pytest.fixture
def logger(caplog: pytest.LogCaptureFixture):
    with caplog.at_level(logging.INFO, "test_logger"):
        yield get_logger("test_logger")


def test_add_logging_context(logger: logging.Logger, caplog: pytest.LogCaptureFixture):
    """Test that :py:class:`add_logging_context` works."""
    logger.info("1")
    records = cast(List[Any], caplog.records)
    assert records[-1].message == "1"

    with add_logging_context(a=1, k={"hello": "world"}):
        logger.info("2")
        assert records[-1].message == "2"
        assert records[-1].a == 1
        assert records[-1].k == {"hello": "world"}

        with add_logging_context(b=2, k={"foo": "bar"}):
            logger.info("3")
            assert records[-1].message == "3"
            assert records[-1].a == 1
            assert records[-1].b == 2
            assert records[-1].k == {"hello": "world", "foo": "bar"}

            with add_logging_context(a=4, c=3, k={"hello": "world2", "biz": "baz"}):
                logger.info("4")
                assert records[-1].message == "4"
                assert records[-1].a == 4
                assert records[-1].b == 2
                assert records[-1].c == 3
                assert records[-1].k == {"hello": "world2", "foo": "bar", "biz": "baz"}

            logger.info("5")
            assert records[-1].message == "5"
            assert records[-1].a == 1
            assert records[-1].b == 2
            assert records[-1].k == {"hello": "world", "foo": "bar"}
            assert not hasattr(records[-1], "c")

        logger.info("6")
        assert records[-1].message == "6"
        assert records[-1].a == 1
        assert records[-1].k == {"hello": "world"}
        assert not hasattr(records[-1], "b")
        assert not hasattr(records[-1], "c")

    logger.info("7")
    assert records[-1].message == "7"
    assert not hasattr(records[-1], "a")
    assert not hasattr(records[-1], "b")
    assert not hasattr(records[-1], "c")


def test_inline_extra(logger: logging.Logger, caplog: pytest.LogCaptureFixture):
    """Test that we can add one-off additions to our context."""
    logger.info("message", extra={"hello": "world"})
    records = cast(List[Any], caplog.records)
    assert records[-1].message == "message"
    assert records[-1].hello == "world"

    with add_logging_context(foo="bar"):
        logger.info("message", extra={"hello": "world"})
        assert records[-1].message == "message"
        assert records[-1].hello == "world"
        assert records[-1].foo == "bar"


def test_thread_local(logger: logging.Logger, caplog: pytest.LogCaptureFixture):
    """Test that our logging context is indeed thread-local."""

    def thread_worker(val: int):
        records = cast(List[Any], caplog.records)
        logger.info("1")
        assert records[-1].message == "1"
        assert not hasattr(records[-1], "a")

        with add_logging_context(a=2, b=val):
            logger.info("2")
            assert records[-1].message == "2"
            assert records[-1].a == 2
            assert records[-1].b == val

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as exc:
        with add_logging_context(a=1):
            list(exc.map(thread_worker, [1, 2]))



from __future__ import annotations

import enum
import time
from datetime import datetime
from itertools import chain
from typing import TYPE_CHECKING, Any, Iterable, List, Mapping, Optional, Sequence, Type, TypeVar, Union
from urllib.parse import urljoin

import pandas as pd
import pyarrow as pa
import requests
from pydantic import BaseModel, ValidationError
from requests import HTTPError

from chalk.client.client_protocol import (
    ChalkAPIClientProtocol,
    ChalkBaseException,
    ChalkError,
    FeatureResult,
    OfflineQueryContext,
    OnlineQueryContext,
    OnlineQueryResponse,
    ResolverRunResponse,
    WhoAmIResponse,
)
from chalk.client.dataset import DatasetVersion, load_dataset
from chalk.config.auth_config import load_token
from chalk.features import DataFrame, Feature, Features, FeatureWrapper, ensure_feature, unwrap_feature
from chalk.features.pseudofeatures import CHALK_TS_FEATURE
from chalk.serialization.codec import FEATURE_CODEC
from chalk.utils.log_with_context import get_logger

if TYPE_CHECKING:
    import polars as pl


_logger = get_logger(__name__)


class _ExchangeCredentialsRequest(BaseModel):
    client_id: str
    client_secret: str
    grant_type: str


class _ExchangeCredentialsResponse(BaseModel):
    access_token: str
    token_type: str
    expires_in: int
    engines: Optional[Mapping[str, str]] = None


class _OfflineQueryResponse(BaseModel):
    columns: List[str]
    output: List[List[Any]]
    errors: Optional[List[ChalkError]]


class _OfflineQueryInput(BaseModel):
    columns: List[str]
    values: List[List[Any]]


class _QueryRequest(BaseModel):
    inputs: Mapping[str, Any]
    outputs: List[str]
    staleness: Optional[Mapping[str, str]] = None
    context: Optional[OnlineQueryContext]
    deployment_id: Optional[str] = None
    correlation_id: Optional[str] = None
    query_name: Optional[str] = None
    meta: Optional[Mapping[str, str]] = None


class _TriggerResolverRunRequest(BaseModel):
    resolver_fqn: str


T = TypeVar("T")


class _ChalkClientConfig(BaseModel):
    client_id: str
    client_secret: str
    api_server: str
    active_environment: Optional[str]


class _OnlineQueryResponse(BaseModel):
    data: List[FeatureResult]
    errors: Optional[List[ChalkError]]


class _CreateOfflineQueryJobRequest(BaseModel):
    """
    Attributes
        output: A list of output feature root fqns to query
        destination_format: The desired output format. Should be 'CSV' or 'PARQUET'
        input: Any givens
        max_samples: The maximum number of samples
    """

    output: List[str]
    destination_format: str
    input: Optional[_OfflineQueryInput] = None
    max_samples: Optional[int] = None
    observed_at_lower_bound: Optional[datetime] = None
    observed_at_upper_bound: Optional[datetime] = None
    dataset_name: Optional[str] = None
    branch: Optional[str] = None
    environment: Optional[str] = None


class _CreateOfflineQueryJobResponse(BaseModel):
    """
    Attributes:
        job_id: A job ID, which can be used to retrieve the results.
    """

    job_id: str
    version: int = 1  # Deprecated
    errors: Optional[List[ChalkError]]


class _GetOfflineQueryJobResponse(BaseModel):
    """
    Attributes:
        is_finished: Whether the export job is finished (it runs asynchronously)
        signed_urls: A list of signed URLs that the client can download to retrieve the exported data.
    """

    is_finished: bool
    version: int
    urls: List[str]


class ChalkOfflineQueryException(ChalkBaseException):
    message: str
    errors: List[ChalkError]

    def __init__(self, message: str, errors: List[ChalkError]):
        self.message = message
        self.errors = errors
        super().__init__(message + "\n" + "\n".join(["\t" + e.message for e in errors[0:3]]))


class ChalkResolverRunException(ChalkBaseException):
    message: str

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class ChalkDatasetDownloadException(ChalkBaseException):
    message: str

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class OnlineQueryResponseWrapper(OnlineQueryResponse):
    data: List[FeatureResult]
    errors: Optional[List[ChalkError]]

    def __init__(
        self,
        data: List[FeatureResult],
        errors: Optional[List[ChalkError]],
    ):
        self.data = data
        self.errors = errors
        for d in self.data:
            if d.value is not None:
                d.value = FEATURE_CODEC.decode_fqn(d.field, d.value)
        self._values = {d.field: d for d in self.data}

    def get_feature(self, feature: Any) -> Optional[FeatureResult]:
        # Typing `feature` as Any, as the Features will be typed as the underlying datatypes, not as Feature
        return self._values.get(ensure_feature(feature).root_fqn)

    def get_feature_value(self, feature: Any) -> Optional[Any]:
        # Typing `feature` as Any, as the Features will be typed as the underlying datatypes, not as Feature
        v = self.get_feature(feature)
        return v and v.value


def _expand_scalar_features_shallow(o: Union[Feature, FeatureWrapper, Features, str, Any]) -> Iterable[Feature]:
    if isinstance(o, str):
        yield Feature.from_root_fqn(o)
        return
    if isinstance(o, type) and issubclass(o, Features):
        for f in o.features:
            if f.is_scalar:
                yield f
        return
    if isinstance(o, FeatureWrapper):
        o = unwrap_feature(o)
    if isinstance(o, Feature):
        if o.is_scalar:
            yield o
            return
        if o.is_has_one:
            assert o.joined_class is not None
            for f in o.joined_class.features:
                if isinstance(f, Feature) and f.is_scalar:
                    yield Feature.from_root_fqn(f"{o.root_fqn}.{f.name}")
            return
    raise ValueError(f"Unsupported feature {o}")


class ChalkAPIClientImpl(ChalkAPIClientProtocol):
    def __init__(
        self,
        *,
        client_id: Optional[str],
        client_secret: Optional[str],
        environment: Optional[str],
        api_server: Optional[str],
    ):
        if client_id is not None and client_secret is not None:
            self._config = _ChalkClientConfig(
                client_id=client_id,
                client_secret=client_secret,
                api_server=api_server or "https://api.chalk.ai",
                active_environment=environment,
            )
        else:
            token = load_token()
            if token is None:
                raise ValueError(
                    (
                        "Could not find .chalk.yml config file for project, "
                        "and explicit configuration was not provided. "
                        "You may need to run `chalk login` from your command line, "
                        "or check that your working directory is set to the root of "
                        "your project."
                    )
                )
            self._config = _ChalkClientConfig(
                client_id=token.clientId,
                client_secret=token.clientSecret,
                api_server=api_server or token.apiServer or "https://api.chalk.ai",
                active_environment=environment or token.activeEnvironment,
            )

        self._default_headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        self._engines: Optional[Mapping[str, str]] = None
        self._exchanged_credentials = False

    def _exchange_credentials(self):
        _logger.debug("Performing a credentials exchange")
        resp = requests.post(
            url=urljoin(self._config.api_server, f"v1/oauth/token"),
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
            json=_ExchangeCredentialsRequest(
                client_id=self._config.client_id,
                client_secret=self._config.client_secret,
                grant_type="client_credentials",
            ).dict(),
            timeout=10,
        )
        resp.raise_for_status()
        response_json = resp.json()
        try:
            creds = _ExchangeCredentialsResponse(**response_json)
        except ValidationError:
            raise HTTPError(response=resp)
        self._default_headers["Authorization"] = f"Bearer {creds.access_token}"
        self._engines = creds.engines
        self._exchanged_credentials = True

    def _get_headers(self, environment_override: Optional[str]):
        x_chalk_env_id = environment_override or self._config.active_environment
        headers = dict(self._default_headers)  # shallow copy
        if x_chalk_env_id is not None:
            headers["X-Chalk-Env-Id"] = x_chalk_env_id
        return headers

    def _request(
        self,
        method: str,
        uri: str,
        response: Type[T],
        json: Optional[BaseModel] = None,
        use_engine: bool = False,
        environment_override: Optional[str] = None,
    ) -> T:
        # Track whether we already exchanged credentials for this request
        exchanged_credentials = False
        if not self._exchanged_credentials:
            exchanged_credentials = True
            self._exchange_credentials()
        headers = self._get_headers(environment_override=environment_override)
        active_env = environment_override or self._config.active_environment
        if use_engine and self._engines is not None and active_env in self._engines:
            assert isinstance(active_env, str)
            base = self._engines[active_env]
        else:
            base = self._config.api_server
        url = urljoin(base, uri)
        json_body = json and json.dict()
        r = requests.request(method=method, headers=headers, url=url, json=json_body)
        if r.status_code in (401, 403) and not exchanged_credentials:
            # It is possible that credentials expired, or that we changed permissions since we last
            # got a token. Exchange them and try again
            self._exchange_credentials()
            r = requests.request(method=method, headers=headers, url=url, json=json_body)

        r.raise_for_status()
        return response(**r.json())

    def whoami(self) -> WhoAmIResponse:
        return self._request(method="GET", uri=f"/v1/who-am-i", response=WhoAmIResponse)

    def upload_features(
        self,
        input: Mapping[Union[str, Feature, Any], Any],
        context: Optional[OnlineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        correlation_id: Optional[str] = None,
        query_name: Optional[str] = None,
        meta: Optional[Mapping[str, str]] = None,
    ) -> Optional[List[ChalkError]]:
        return self.query(
            input=input,
            output=list(input.keys()),
            staleness=None,
            context=context,
            preview_deployment_id=preview_deployment_id,
            correlation_id=correlation_id,
            query_name=query_name,
            meta=meta,
        ).errors

    def query(
        self,
        input: Mapping[Union[str, Feature, FeatureWrapper, Any], Any],
        output: Sequence[Union[str, Feature, FeatureWrapper, Any]],
        staleness: Optional[Mapping[Union[str, Feature, FeatureWrapper, Any], str]] = None,
        context: Optional[OnlineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        correlation_id: Optional[str] = None,
        query_name: Optional[str] = None,
        meta: Optional[Mapping[str, str]] = None,
    ) -> OnlineQueryResponse:
        encoded_inputs = {}
        for feature, value in input.items():
            feature = ensure_feature(feature)
            encoded_inputs[feature.fqn] = FEATURE_CODEC.encode(feature, value)
        outputs: List[str] = []
        for x in output:
            outputs.extend(f.root_fqn for f in _expand_scalar_features_shallow(x))
        request = _QueryRequest(
            inputs=encoded_inputs,
            outputs=outputs,
            staleness={} if staleness is None else {ensure_feature(k).root_fqn: v for k, v in staleness.items()},
            context=context,
            deployment_id=preview_deployment_id,
            correlation_id=correlation_id,
            query_name=query_name,
            meta=meta,
        )

        resp = self._request(
            method="POST",
            uri="/v1/query/online",
            json=request,
            response=_OnlineQueryResponse,
            use_engine=preview_deployment_id is None,
            environment_override=context.environment if context else None,
        )
        return OnlineQueryResponseWrapper(
            data=resp.data,
            errors=resp.errors,
        )

    def get_training_dataframe(
        self,
        input: Union[Mapping[Union[str, Feature], Any], pl.DataFrame, pd.DataFrame, DataFrame],
        input_times: List[datetime],
        output: List[Union[str, Feature, Any]],
        context: Optional[OfflineQueryContext] = None,
        dataset: Optional[str] = None,
        branch: Optional[str] = None,
    ) -> pd.DataFrame:
        if not isinstance(input, DataFrame):
            input = DataFrame(input)

        self._get_headers(environment_override=None)

        return self._get_training_dataframe(
            input=input.to_pandas(),
            input_times=input_times,
            output=output,
            context=context,
            dataset=dataset,
            branch=branch,
        )

    def _decode_offline_response(self, offline_query_response: _OfflineQueryResponse) -> pd.DataFrame:
        data = {}
        for col_index, column in enumerate(offline_query_response.output):
            series_values = []
            for value in column:
                value = FEATURE_CODEC.decode_fqn(
                    fqn=offline_query_response.columns[col_index],
                    value=value,
                )
                if isinstance(value, enum.Enum):
                    value = value.value
                series_values.append(value)

            data[offline_query_response.columns[col_index]] = pd.Series(
                data=series_values,
                dtype=FEATURE_CODEC.get_pandas_dtype(offline_query_response.columns[col_index]),
            )
        return pd.DataFrame(data)

    def _get_training_dataframe(
        self,
        input: pd.DataFrame,
        input_times: List[datetime],
        output: List[Union[str, Feature, Any]],
        context: Optional[OfflineQueryContext] = None,
        dataset: Optional[str] = None,
        branch: Optional[str] = None,
    ) -> pd.DataFrame:
        columns = input.columns
        matrix = input.T.values.tolist()

        columns_fqn = [ensure_feature(c).root_fqn for c in chain(columns, (CHALK_TS_FEATURE,))]

        matrix.append([a for a in input_times])

        for col_index, column in enumerate(matrix):
            for row_index, value in enumerate(column):
                matrix[col_index][row_index] = FEATURE_CODEC.encode_fqn(
                    fqn=columns_fqn[col_index],
                    value=value,
                )

        query_input = _OfflineQueryInput(
            columns=columns_fqn,
            values=matrix,
        )

        query_output: List[Feature] = []
        for x in output:
            query_output.extend(_expand_scalar_features_shallow(x))

        try:
            response = self._create_and_await_offline_query_job(
                output=query_output,
                input=query_input,
                dataset_name=dataset,
                branch=branch,
                environment=context and context.environment,
            )
            return response.to_pandas()
        except HTTPError as e:
            _logger.debug("Got HTTP Exception while processing sample query", exc_info=e)
            raise ChalkOfflineQueryException(message=f"HTTP error while processing sample query.", errors=[])

    def sample(
        self,
        output: List[Union[str, Feature, FeatureWrapper, Any]],
        max_samples: Optional[int] = None,
        context: Optional[OfflineQueryContext] = None,
    ) -> pd.DataFrame:
        query_output: list[Feature] = []
        for x in output:
            query_output.extend(_expand_scalar_features_shallow(x))

        try:
            response = self._create_and_await_offline_query_job(output=query_output, max_samples=max_samples)
        except HTTPError as e:
            _logger.debug("Got HTTP Exception while processing sample query", exc_info=e)
            raise ChalkOfflineQueryException(message=f"HTTP error while processing sample query.", errors=[])

        return response.to_pandas()

    def trigger_resolver_run(
        self,
        resolver_fqn: str,
        deployment_id: Optional[str] = None,
    ) -> ResolverRunResponse:
        _logger.debug(f'Triggering resolver {resolver_fqn} to run with deployment ID "{deployment_id}"')

        request = _TriggerResolverRunRequest(resolver_fqn=resolver_fqn)
        try:
            response = self._request(
                method="POST",
                uri="/v1/runs/trigger",
                json=request,
                response=ResolverRunResponse,
            )
        except HTTPError as e:
            message = str(e)

            detail = e.response.json().get("detail")
            if detail is not None:
                message = detail

            raise ChalkResolverRunException(message=message)

        return response

    def get_run_status(self, run_id: str) -> ResolverRunResponse:
        try:
            response = self._request(
                method="GET",
                uri=f"/v1/runs/{run_id}",
                response=ResolverRunResponse,
            )
        except HTTPError as e:
            message = str(e)

            detail = e.response.json().get("detail")
            if detail is not None:
                message = detail

            raise ChalkResolverRunException(message=message)

        return response

    def _create_and_await_offline_query_job(
        self,
        output: Sequence[Feature],
        input: Optional[_OfflineQueryInput] = None,
        max_samples: Optional[int] = None,
        dataset_name: Optional[str] = None,
        branch: Optional[str] = None,
        environment: Optional[str] = None,
    ) -> pl.DataFrame:
        output = list(output)
        req = _CreateOfflineQueryJobRequest(
            output=[x.root_fqn for x in output],
            destination_format="PARQUET",
            input=input,
            max_samples=max_samples,
            dataset_name=dataset_name,
            branch=branch,
            environment=environment,
        )
        response = self._create_offline_query_job(request=req)

        if response.errors is not None and len(response.errors) > 0:
            raise ChalkOfflineQueryException(message="Failed to execute offline query", errors=response.errors)

        while True:
            status = self._get_job_status(job_id=response.job_id)
            if status.is_finished:
                break
            time.sleep(0.5)
        return load_dataset(
            uris=status.urls,
            version=DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES,
            features=output,
        )

    def _create_offline_query_job(self, request: _CreateOfflineQueryJobRequest):
        response = self._request(
            method="POST", uri="/v2/offline_query", json=request, response=_CreateOfflineQueryJobResponse
        )
        return response

    def _get_job_status(self, job_id: str) -> _GetOfflineQueryJobResponse:
        return self._request(method="GET", uri=f"/v2/offline_query/{job_id}", response=_GetOfflineQueryJobResponse)



"""
Adds thread-local context to a Python logger. Taken from neocrym/log-with-context
"""
import collections.abc
import contextlib
import contextvars
import logging
from typing import Any, List, Mapping, Optional

_LOGGING_CONTEXT_STACK: contextvars.ContextVar[List[Mapping[str, Any]]] = contextvars.ContextVar(
    "_LOGGING_CONTEXT_STACK"
)


def _recursive_merge(a: Mapping[str, Any], b: Mapping[str, Any]):
    ans: Mapping[str, Any] = {**a}
    for k, v in b.items():
        if k not in ans or not (isinstance(ans[k], collections.abc.Mapping) and isinstance(v, collections.abc.Mapping)):
            ans[k] = v
            continue
        ans[k] = _recursive_merge(ans[k], v)
    return ans


def get_logging_context() -> Mapping[str, Any]:
    """
    Retrieve the log context for the current python context.
    This initializes the thread-local variable if necessary.
    """
    return _LOGGING_CONTEXT_STACK.get([{}])[-1]


def set_logging_context(_merge: bool = True, **extra: Any):
    """
    Sets the context-local log context to a new value.

    Args:
        merge: If true, the new context will be merged with the current context. Otherwise, the new context will overwrite the existing context.
    """
    log_context_stack = _LOGGING_CONTEXT_STACK.get([{}])
    if _merge:
        extra = _recursive_merge(
            log_context_stack[-1],
            extra,
        )
    log_context_stack.append(extra)
    _LOGGING_CONTEXT_STACK.set(log_context_stack)


def pop_logging_context():
    """Pop the latest log context off of the log context stack, and reset it to its previous value.

    Raises:
        RuntimeError, if there is no current context from set_logging_context()
    """
    stack = _LOGGING_CONTEXT_STACK.get([])
    if len(stack) == 0:
        raise RuntimeError("Log context stack is empty")
    stack.pop()
    _LOGGING_CONTEXT_STACK.set(stack)


class LogWithContextFilter(logging.Filter):
    """Filter to append the ``extras`` onto the LogRecord."""

    def filter(self, record: logging.LogRecord) -> bool:
        for k, v in get_logging_context().items():
            if not hasattr(record, k):
                setattr(record, k, v)
        return True


def get_logger(name: Optional[str]):
    logger = logging.getLogger(name)
    logger.addFilter(LogWithContextFilter())
    return logger


# Backwards compatibility
Logger = get_logger


@contextlib.contextmanager
def add_logging_context(_merge: bool = True, **extra: Any):
    """
    A context manager to push and pop "extra" dictionary keys.

    Args:
        _merge: Whether to merge the new context with the existing log context.
        kwargs: Contextual information to add to the log record
    """
    set_logging_context(_merge, **extra)
    try:
        yield
    finally:
        pop_logging_context()



import dataclasses
import json
from typing import Any, List, Type, TypeVar

import pydantic
from pydantic.main import BaseModel

from chalk.features import DataFrame, Feature, Filter
from chalk.features.feature_field import HasOnePathObj
from chalk.features.resolver import Cron, OfflineResolver, OnlineResolver, SinkResolver, StreamResolver
from chalk.parsed.duplicate_input_gql import (
    UpsertDataFrameGQL,
    UpsertFeatureGQL,
    UpsertFeatureIdGQL,
    UpsertFeatureReferenceGQL,
    UpsertFeatureTimeKindGQL,
    UpsertFilterGQL,
    UpsertHasManyKindGQL,
    UpsertHasOneKindGQL,
    UpsertReferencePathComponentGQL,
    UpsertResolverGQL,
    UpsertResolverOutputGQL,
    UpsertScalarKindGQL,
    UpsertSinkResolverGQL,
    UpsertStreamResolverGQL,
    UpsertStreamResolverParamGQL,
    UpsertStreamResolverParamKeyedStateGQL,
    UpsertStreamResolverParamMessageGQL,
)
from chalk.streams import Windowed
from chalk.streams.types import (
    StreamResolverParam,
    StreamResolverParamKeyedState,
    StreamResolverParamMessage,
    StreamResolverParamMessageWindow,
)
from chalk.utils import paths

T = TypeVar("T")


try:
    import attrs
except ImportError:
    attrs = None


@dataclasses.dataclass
class _ConvertedType:
    name: str
    bases: List[str]


def _get_qualified_class_name(cls: Type[Any]):
    mod = cls.__module__
    return cls.__qualname__ if mod == "builtins" else f"{mod}.{cls.__qualname__}"


def _convert_type(underlying_type: Type) -> _ConvertedType:
    if isinstance(underlying_type, Windowed):
        return _ConvertedType(name=f"Windowed[{str(underlying_type._kind)}]", bases=["CHALK_WINDOWED"])
    assert isinstance(underlying_type, type), f"Underlying type is not a type: '{str(underlying_type)}'"
    base_classes = [x.__name__ for x in type.mro(underlying_type)]

    # Attrs and Dataclasses don't technically have base classes
    # Pydantic calls their base class BaseModel which is way too generic for string comparison
    # For simplicity on the server-side validation, we'll come up with our own "base class" names
    if dataclasses.is_dataclass(underlying_type):
        base_classes.append("__dataclass__")
    if attrs is not None and isinstance(underlying_type, type) and attrs.has(underlying_type):
        base_classes.append("__attrs__")
    if issubclass(underlying_type, pydantic.BaseModel):
        base_classes.append("__pydantic__")

    return _ConvertedType(
        name=underlying_type.__name__,
        bases=base_classes,
    )


def _get_feature_id(s: Feature):
    assert s.name is not None, "Could not find feature name"
    assert s.namespace is not None, "Could not find feature namespace"
    return UpsertFeatureIdGQL(
        fqn=s.fqn,
        name=s.name,
        namespace=s.namespace,
    )


def _convert_df(df: Type[DataFrame]) -> UpsertDataFrameGQL:
    return UpsertDataFrameGQL(
        columns=[_get_feature_id(f) for f in df.columns if not f.is_autogenerated],
        filters=[convert_type_to_gql(f) for f in df.filters],
    )


def _get_path_component(pc: HasOnePathObj) -> UpsertReferencePathComponentGQL:
    assert isinstance(pc.parent, Feature), f"Parent in relationship path not a feature, but {type(pc).__name__}"
    return UpsertReferencePathComponentGQL(
        parent=_get_feature_id(pc.parent),
        child=_get_feature_id(pc.child),
        parentToChildAttributeName=pc.parent_to_child_attribute_name,
    )


def convert_type_to_gql(t: Any):
    if isinstance(t, StreamResolver):
        return UpsertStreamResolverGQL(
            fqn=t.fqn,
            kind="stream",
            sourceClassName=_get_qualified_class_name(t.source.__class__),
            sourceConfig=t.source.config_to_json(),
            functionDefinition=t.function_definition,
            environment=[t.environment] if isinstance(t.environment, str) else t.environment,
            doc=t.fn.__doc__,
            machineType=t.machine_type,
            output=[
                _get_feature_id(f)
                for f in t.output_features
                if (not isinstance(f, type) or not issubclass(f, DataFrame)) and not f.is_autogenerated
            ],
            inputs=[
                convert_type_to_gql(i)
                for i in t.signature.params
                if not isinstance(i, StreamResolverParamMessageWindow)
            ],
        )

    if isinstance(t, SinkResolver):
        return UpsertSinkResolverGQL(
            fqn=t.fqn,
            functionDefinition=t.function_definition,
            inputs=[
                UpsertFeatureReferenceGQL(
                    underlying=_get_feature_id(f),
                    path=[_get_path_component(p) for p in f.path or []],
                )
                for f in t.inputs
                if not f.is_autogenerated
            ],
            environment=t.environment,
            tags=t.tags,
            doc=t.doc,
            machineType=t.machine_type,
            bufferSize=t.buffer_size,
            debounce=t.debounce,
            maxDelay=t.max_delay,
            upsert=t.upsert,
        )

    if isinstance(t, (OnlineResolver, OfflineResolver)):
        cron = t.cron
        if isinstance(cron, Cron):
            assert t.cron.schedule is not None, "`Cron` object must be constructed with a `schedule` property."
            cron = cron.schedule
        return UpsertResolverGQL(
            fqn=t.fqn,
            kind="offline" if isinstance(t, OfflineResolver) else "online",
            functionDefinition=t.function_definition,
            inputs=[
                UpsertFeatureReferenceGQL(
                    underlying=_get_feature_id(f),
                    path=[_get_path_component(p) for p in f.path or []],
                )
                for f in t.inputs
                if not f.is_autogenerated
            ],
            output=UpsertResolverOutputGQL(
                features=[
                    _get_feature_id(f)
                    for f in t.output.features
                    if (not isinstance(f, type) or not issubclass(f, DataFrame)) and not f.is_autogenerated
                ],
                dataframes=[
                    _convert_df(f) for f in t.output.features if isinstance(f, type) and issubclass(f, DataFrame)
                ],
            ),
            environment=t.environment,
            tags=t.tags,
            doc=t.doc,
            cron=cron,
            machineType=t.machine_type,
        )

    if isinstance(t, StreamResolverParam):
        if isinstance(t, StreamResolverParamMessage):
            converted_type = _convert_type(t.typ)
            schema = None
            if issubclass(t.typ, BaseModel):
                schema = json.loads(t.typ.schema_json())
            return UpsertStreamResolverParamGQL(
                message=UpsertStreamResolverParamMessageGQL(
                    name=t.name,
                    typeName=converted_type.name,
                    bases=converted_type.bases,
                    schema=schema,
                ),
                state=None,
            )
        elif isinstance(t, StreamResolverParamKeyedState):
            converted_type = _convert_type(t.typ)
            default_value = None
            schema = None
            if dataclasses.is_dataclass(t.default_value):
                default_value = json.loads(json.dumps(dataclasses.asdict(t.default_value)))
                schema = pydantic.dataclasses.dataclass(t.typ).__pydantic_model__.schema()
            elif isinstance(t.default_value, BaseModel):
                default_value = json.loads(t.default_value.json())
                schema = json.loads(t.typ.schema_json())
            return UpsertStreamResolverParamGQL(
                state=UpsertStreamResolverParamKeyedStateGQL(
                    name=t.name,
                    defaultValue=default_value,
                    typeName=converted_type.name,
                    bases=converted_type.bases,
                    schema=schema,
                ),
                message=None,
            )

    if isinstance(t, Feature):
        assert t.name is not None, "Feature has no name"
        assert t.namespace is not None, "Feature has no namespace"
        scalar_kind_gql = None
        has_one_kind_gql = None
        has_many_kind_gql = None
        feature_time_kind_gql = None
        if t.is_autogenerated:
            raise RuntimeError("Autogenerated features should not be converted")
        if t.is_has_one:
            has_one_kind_gql = UpsertHasOneKindGQL(join=convert_type_to_gql(t.join))
        elif t.is_has_many:
            has_many_kind_gql = UpsertHasManyKindGQL(
                join=convert_type_to_gql(t.join),
                columns=None,
                filters=None,
            )
        elif t.is_feature_time:
            feature_time_kind_gql = UpsertFeatureTimeKindGQL()
        else:
            assert t.typ is not None
            converted_type = _convert_type(t.typ.underlying)
            scalar_kind_gql = UpsertScalarKindGQL(
                scalarKind=converted_type.name,
                primary=t.primary,
                baseClasses=converted_type.bases,
                version=t.version,
                hasEncoderAndDecoder=t.encoder is not None and t.decoder is not None,
            )

        return UpsertFeatureGQL(
            id=UpsertFeatureIdGQL(
                fqn=t.fqn,
                name=t.name,
                namespace=t.namespace,
            ),
            maxStaleness=t.max_staleness,
            description=t.description,
            owner=t.owner,
            windowBuckets=t.window_buckets and list(t.window_buckets),
            etlOfflineToOnline=t.etl_offline_to_online,
            tags=t.tags,
            hasOneKind=has_one_kind_gql,
            hasManyKind=has_many_kind_gql,
            scalarKind=scalar_kind_gql,
            featureTimeKind=feature_time_kind_gql,
            namespacePath=str(paths.get_classpath(t.features_cls)),
        )

    if isinstance(t, Filter):
        return UpsertFilterGQL(
            lhs=_get_feature_id(t.lhs),
            op=t.operation,
            rhs=_get_feature_id(t.rhs),
        )

    return t



import pytest

from chalk.features import features
from chalk.sql import PostgreSQLSource
from chalk.sql._internal.integrations.postgres import PostgreSQLSourceImpl

try:
    import psycopg2
except ImportError:
    psycopg2 = None

try:
    import psycopg2
except ImportError:
    psycopg2 = None


@features
class UserFeatures:
    id: str


@features
class LibraryFeatures:
    id: str


@pytest.mark.skipif(psycopg2 is None, reason="psycopg2 is not installed")
def test_table_builder():
    source = (
        PostgreSQLSource()
        .with_table(name="users", features=UserFeatures)
        .with_table(name="libraries", features=LibraryFeatures)
    )
    assert isinstance(source, PostgreSQLSourceImpl)
    assert len(source.ingested_tables) == 2



import pytest

from chalk.streams import KafkaSource


@pytest.fixture
def source():
    return KafkaSource(topic="topic", bootstrap_server="bootstrap-server")


def test_sasl_mechanism_default(source: KafkaSource):
    assert source.sasl_mechanism == "PLAIN"


def test_security_protocol_default(source: KafkaSource):
    assert source.security_protocol == "PLAINTEXT"


def test_cliend_id_prefixdefault(source: KafkaSource):
    assert source.client_id_prefix == "chalk/"


def test_group_id_default(source: KafkaSource):
    assert source.group_id_prefix == "chalk/"



# AUTO-GENERATED FILE.
# Re-run /chalk/feature_n/codegen.py to regenerate contents.
from typing import TypeVar, Generic, Optional, Dict
from chalk.features.dataframe import DataFrameMeta


T1 = TypeVar("T1")
T2 = TypeVar("T2")


class Features(Generic[T1, T2]):
    pass


class DataFrame(Generic[T1, T2], metaclass=DataFrameMeta):
    def __getitem__(self, item):
        pass



from __future__ import annotations

import collections.abc
import dataclasses
import enum
from datetime import date, datetime, timezone
from typing import TYPE_CHECKING, Any, List, Type, Union, cast, get_type_hints

import cattrs
import pendulum
import pyarrow
import pyarrow as pa
import pydantic
import pytz
from dateutil import parser
from pendulum.tz.timezone import Timezone
from pydantic import BaseModel
from pydantic.fields import ModelField
from typing_extensions import get_args, get_origin

from chalk.serialization.parsed_annotation import ParsedAnnotation
from chalk.streams import Windowed
from chalk.utils.enum import get_enum_value_type
from chalk.utils.log_with_context import get_logger
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    from chalk.features import Feature

try:
    import attrs
except ImportError:
    # Imports not available. Attrs is not required.
    attrs = None

try:
    import pandas as pd
except ImportError:
    pd = None

try:
    import numpy as np
except ImportError:
    np = None

_log = get_logger(__name__)

_NoneType = type(None)


def _unwrap_optional_if_needed(typ: Any):
    if get_origin(typ) == Union and len(get_args(typ)) == 2:
        has_none = any(d == _NoneType for d in get_args(typ))
        if has_none:
            return next(m for m in get_args(typ) if m != _NoneType)

    return typ


def is_numeric(dtype: Union[pl.DataType, Type[pl.DataType]]) -> bool:
    # This function is used only by the engine
    import polars as pl

    return dtype in (
        pl.Float32,
        pl.Float64,
        pl.Int16,
        pl.Int8,
        pl.Int32,
        pl.Int64,
        pl.UInt8,
        pl.UInt16,
        pl.UInt32,
        pl.UInt64,
    )


def _is_namedtuple(value: Any, annotated: bool = False) -> bool:
    """Infer whether value is a NamedTuple."""
    # From https://github.com/pola-rs/polars/blob/5f3e332fb2a653064f083b02949c527e0ec0afda/py-polars/polars/internals/construction.py#L78
    if all(hasattr(value, attr) for attr in ("_fields", "_field_defaults", "_replace")):
        return len(value.__annotations__) == len(value._fields) if annotated else True
    return False


class FeatureCodec:
    def __init__(self):
        self.converter = cattrs.Converter()
        self.converter.register_structure_hook(datetime, lambda v, _: parser.isoparse(v))
        self.converter.register_unstructure_hook(
            datetime,
            lambda v: (
                cast(datetime, v) if cast(datetime, v).tzinfo else pytz.utc.localize(cast(datetime, v))
            ).isoformat(),
        )

    def _default_encode(self, value: Any):
        if isinstance(value, set):
            return {self._default_encode(x) for x in value}
        if isinstance(value, list):
            return [self._default_encode(x) for x in value]
        if isinstance(value, (str, int, float)):
            return value
        if isinstance(value, enum.Enum):
            return self._default_encode(value.value)
        if isinstance(value, datetime):
            tz = value.tzinfo or (datetime.now(timezone.utc).astimezone().tzinfo)
            assert tz is not None
            return pendulum.instance(value, cast(Timezone, tz)).isoformat()
        if isinstance(value, date):
            return value.isoformat()
        if np is not None and isinstance(value, np.integer):
            return int(value)
        if np is not None and isinstance(value, np.floating):
            return float(value)
        if pd is not None and isinstance(value, pd.Timestamp):
            return pendulum.instance(value.to_pydatetime()).isoformat()
        if isinstance(value, pydantic.BaseModel):
            return value.dict()
        if (attrs is not None and attrs.has(type(value))) or dataclasses.is_dataclass(value):
            return self.converter.unstructure(value)
        if isinstance(value, collections.abc.Iterable):
            return [self._default_encode(x) for x in value]
        raise TypeError(f"Unable to encode value of type {type(value).__name__}")

    def encode(
        self,
        feature: Feature,
        value: Any,
    ):
        if value is None:
            return None

        if feature.encoder is not None:
            return feature.encoder(value)

        return self._default_encode(value)

    def encode_fqn(self, fqn: str, value: Any):
        from chalk.features import Feature

        return self.encode(Feature.from_root_fqn(fqn), value)

    def _to_polars_struct(self, fqn: str, value: BaseModel):
        try:
            import polars as pl
        except ImportError:
            raise missing_dependency_exception("chalkpy[polars]")

        return pl.DataFrame(
            data={
                field.alias or field.name: [getattr(value, field.alias or field.name)]
                for field in value.__fields__.values()
            },
            columns={
                field.alias
                or field.name: self._pa_to_pl(self._get_pa_dtype_for_type(field.type_, field.alias or field.name))
                for field in value.__fields__.values()
            },
        ).to_struct(fqn)

    def _default_decode_value(
        self,
        feature: Feature,
        value: Any,
    ):
        assert feature.typ is not None
        if isinstance(value, feature.typ.underlying):
            return value
        if issubclass(feature.typ.underlying, enum.Enum):
            value = feature.typ.underlying(value)
        if issubclass(feature.typ.underlying, datetime):
            value = parser.isoparse(value)
        elif issubclass(feature.typ.underlying, date):
            # note: datetime is a subclass of date, so we must be careful to decode accordingly
            value = date.fromisoformat(value)
        if issubclass(feature.typ.underlying, pydantic.BaseModel):
            return feature.typ.underlying(**value)
        elif (attrs is not None and attrs.has(feature.typ.underlying)) or dataclasses.is_dataclass(
            feature.typ.underlying
        ):
            return self.converter.structure(value, feature.typ.underlying)
        if not isinstance(value, feature.typ.underlying):
            raise TypeError(
                f"Unable to decode value {value} to type {feature.typ.underlying.__name__} for '{feature.root_fqn}'"
            )
        return value

    def _default_decode(
        self,
        feature: Feature,
        value: Any,
    ):
        assert feature.typ is not None
        if value is None:
            if feature.typ.is_nullable:
                return None
            else:
                raise ValueError(f"Value is none but feature {feature} is not nullable")
        if feature.typ.collection_type is not None and get_origin(feature.typ.collection_type) == set:
            if not isinstance(value, set):
                raise TypeError(f"Feature {feature} is a set but value {value} is not")
            return {self._default_decode_value(feature, x) for x in value}
        elif feature.typ.collection_type is not None and get_origin(feature.typ.collection_type) == list:
            if not isinstance(value, list):
                raise TypeError(f"Feature {feature} is a list but value {value} is not")
            return [self._default_decode_value(feature, x) for x in value]
        else:
            return self._default_decode_value(feature, value)

    def decode(
        self,
        feature: Feature,
        value: Any,
    ):
        if value is None:
            return None

        if feature.decoder is not None:
            return feature.decoder(value)

        return self._default_decode(feature, value)

    def decode_fqn(
        self,
        fqn: str,
        value: Any,
    ):
        from chalk.features import Feature

        try:
            f = Feature.from_root_fqn(fqn)
        except:
            return value

        return self.decode(f, value)

    def get_pandas_dtype(self, fqn: str) -> str:
        from chalk.features import Feature

        feature = Feature.from_root_fqn(fqn)
        typ = feature.typ
        assert typ is not None, "typ should be specified"
        underlying = typ.underlying
        if issubclass(underlying, enum.Enum):
            # For enums, require all members to have the same type
            underlying = get_enum_value_type(underlying)
        # See https://pandas.pydata.org/docs/user_guide/basics.html#basics-dtypes
        if issubclass(underlying, str):
            return "string"
        if issubclass(underlying, bool):
            return "boolean"
        if issubclass(underlying, int):
            return "Int64"
        if issubclass(underlying, float):
            return "Float64"
        if issubclass(underlying, datetime):
            # This assumes timezone-aware. For timezone-unaware the `pandas_dtype` must be set directly on the Feature
            return "datetime64[ns, utc]"
        _log.info(
            f"Defaulting to pandas type 'object' for fqn {fqn} of type {typ.underlying.__name__}. Set the `pandas_dtype` attribute for better specificity."
        )
        return "object"

    def get_polars_dtype(self, fqn: str) -> Union[pl.DataType, Type[pl.DataType]]:
        try:
            import polars as pl
        except ImportError:
            raise missing_dependency_exception("chalkpy[polars]")

        # BEGIN This should be replaced with a structification
        from chalk.features import Feature

        feature = Feature.from_root_fqn(fqn)
        # END structification workaround

        if feature.is_has_one or feature.is_has_many:
            return pl.Object
        pa_type = self.get_pyarrow_dtype(fqn)

        try:
            return self._pa_to_pl(pa_type)
        except Exception as exc:
            raise ValueError(
                f"Failed to determined type for feature {fqn}. Please set the pyarrow_dtype attribute directly."
            ) from exc

    def _pa_to_pl(self, pa_type: pa.DataType) -> Union[pl.DataType, Type[pl.DataType]]:
        try:
            import polars as pl
        except ImportError:
            raise missing_dependency_exception("chalkpy[polars]")

        """May throw"""
        # polars coerces pa.Date to pl.DateTime
        if pa_type == pa.date64():
            return pl.Date

        # To convert the dtypes, we'll create an empty pyarrow table, and then convert that to polars, and get the dtype
        pa_table = pa.Table.from_arrays(arrays=[[]], schema=pa.schema(fields={"col": pa_type}))

        df = pl.from_arrow(pa_table)
        assert isinstance(df, pl.DataFrame)
        return df.dtypes[0]

    def get_pyarrow_dtype(self, fqn: str) -> pa.DataType:
        from chalk.features import Feature

        feature = Feature.from_root_fqn(fqn)
        if feature.pyarrow_dtype is not None:
            return feature.pyarrow_dtype
        typ = feature.typ
        assert typ is not None, "typ should be specified"
        underlying = typ.underlying
        dtype = self._get_pa_dtype_for_type(underlying, fqn)
        if typ.collection_type is not None:
            dtype = pa.list_(dtype)
        return dtype

    def _get_pa_dtype_for_type(self, underlying: type | Windowed, fqn: str) -> pa.DataType:
        # Polars seems to allow optional for any dtype, so we ignore it when computing dtypes
        underlying = _unwrap_optional_if_needed(underlying)
        if isinstance(underlying, Windowed):
            # If the parent window feature appears in a dataframe, then it's likely in a streaming resolver
            # Use the underlying type for that windowed feature
            return self._get_pa_dtype_for_type(underlying._kind, fqn)
        if issubclass(underlying, enum.Enum):
            # For enums, require all members to have the same type
            underlying = get_enum_value_type(underlying)
        if issubclass(underlying, str):
            return pa.string()
        if issubclass(underlying, bool):
            return pa.bool_()
        if issubclass(underlying, int):
            return pa.int64()
        if issubclass(underlying, float):
            return pa.float64()
        if issubclass(underlying, datetime):
            return pa.timestamp("us", "UTC")
        if issubclass(underlying, date):
            return pa.date64()
        if dataclasses.is_dataclass(underlying) or _is_namedtuple(underlying, annotated=True):
            annotations = get_type_hints(underlying)
            fields: List[pa.Field] = []
            for field_name, type_annotation in annotations.items():
                parsed_annotation = ParsedAnnotation(underlying=type_annotation)
                if not parsed_annotation.is_scalar:
                    raise TypeError(f"{fqn} cannot be another Features class ")
                underlying_dtype = self._get_pa_dtype_for_type(parsed_annotation.underlying, f"{fqn}.{field_name}")
                if parsed_annotation.collection_type is None:
                    fields.append(
                        pa.field(
                            field_name,
                            underlying_dtype,
                        )
                    )
                else:
                    fields.append(pa.field(field_name, pa.list_(underlying_dtype)))
            return pa.struct(fields)
        if issubclass(underlying, BaseModel):
            fields: List[pa.Field] = []

            field: ModelField
            for field in underlying.__fields__.values():
                field_name = field.alias if field.alias is not None else field.name
                fields.append(
                    pyarrow.field(
                        metadata={"chalk_field_name": field.name},
                        name=field_name,
                        nullable=field.allow_none,  # noqa
                        # abusing fqn here
                        type=self._get_pa_dtype_for_type(field.type_, f"{fqn}.{field_name}"),
                    )
                )

            return pa.struct(fields)

        raise ValueError(
            f"Unable to determine the PyArrow type for feature {underlying.__name__}. Please set the `pyarrow_dtype` attribute."
        )


FEATURE_CODEC = FeatureCodec()



import contextlib
import io
import os
from typing import Any, Dict, Mapping, Optional, Union

import pyarrow as pa
import pyarrow.csv
from sqlalchemy.engine import URL, Connection

from chalk.integrations.named import load_integration_variable
from chalk.sql._internal.sql_source import BaseSQLSource, TableIngestMixIn
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.sql.protocols import SQLSourceWithTableIngestProtocol
from chalk.utils.environment_parsing import env_var_bool
from chalk.utils.missing_dependency import missing_dependency_exception


class PostgreSQLSourceImpl(BaseSQLSource, TableIngestMixIn, SQLSourceWithTableIngestProtocol):
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[Union[int, str]] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
    ):
        try:
            import psycopg2
        except ImportError:
            raise missing_dependency_exception("chalkpy[postgresql]")
        del psycopg2  # unused
        self.host = host or load_integration_variable(integration_name=name, name="PGHOST")
        self.port = (
            int(port)
            if port is not None
            else load_integration_variable(integration_name=name, name="PGPORT", parser=int)
        )
        self.db = db or load_integration_variable(integration_name=name, name="PGDATABASE")
        self.user = user or load_integration_variable(integration_name=name, name="PGUSER")
        self.password = password or load_integration_variable(integration_name=name, name="PGPASSWORD")
        self.ingested_tables: Dict[str, Any] = {}
        super(PostgreSQLSourceImpl, self).__init__()

    def engine_args(self) -> Mapping[str, Any]:
        return dict(
            pool_size=20,
            max_overflow=60,
            pool_pre_ping=True,
            # Trying to fix mysterious dead connection issue
            connect_args={
                "keepalives": 1,
                "keepalives_idle": 30,
                "keepalives_interval": 10,
                "keepalives_count": 5,
            },
        )

    def local_engine_url(self) -> URL:
        return URL.create(
            drivername="postgresql",
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.db,
        )

    def execute_query(self, finalized_query: FinalizedChalkQuery, connection: Optional[Connection] = None) -> pa.Table:
        import psycopg2.sql

        if env_var_bool("CHALK_DISABLE_POSTGRES_CSV_TO_PYARROW"):
            return super().execute_query(finalized_query, connection)

        stmt, positional_params, named_params = self.compile_query(finalized_query, paramstyle="named")
        assert len(positional_params) == 0, "should not have any positional params"
        # Forcing quote so the unquoted empty string will represent null values
        stmt = f"COPY ({stmt}) TO STDOUT (FORMAT CSV, HEADER true, FORCE_QUOTE *, ESCAPE '\\')"
        # Convert the param style to python3-style {}, which is what psycopg2.sql.SQL.format expects
        for named_param in named_params:
            stmt = stmt.replace(f":{named_param}", ("{" f"{named_param}" "}"))
        formatted_stmt = psycopg2.sql.SQL(stmt).format(
            **{k: psycopg2.sql.Literal(v) for (k, v) in named_params.items()}
        )
        with (self.get_engine().connect() if connection is None else contextlib.nullcontext(connection)) as cnx:
            dbapi = cnx.connection
            with dbapi.cursor() as cursor:
                buffer = io.BytesIO()
                cursor.copy_expert(formatted_stmt, buffer)
        buffer.seek(0)
        return pyarrow.csv.read_csv(
            buffer,
            parse_options=pyarrow.csv.ParseOptions(
                newlines_in_values=True,
                escape_char="\\",
                double_quote=False,
            ),
            convert_options=pyarrow.csv.ConvertOptions(
                true_values=["t"],
                false_values=["f"],
                strings_can_be_null=True,
                quoted_strings_can_be_null=False,
            ),
        )
